import traceback
import warnings

from asdf.tags.core import NDArrayType
import numpy as np
from astropy.modeling.core import Model
from astropy import units as u
from stdatamodels.validate import ValidationWarning

from .reference import ReferenceFileModel

__all__ = ['DistortionModel', 'DistortionMRSModel', 'SpecwcsModel', 'RegionsModel',
           'WavelengthrangeModel', 'CameraModel', 'CollimatorModel', 'OTEModel',
           'FOREModel', "FPAModel", 'IFUPostModel', 'IFUFOREModel', 'IFUSlicerModel',
           'MSAModel', 'FilteroffsetModel', 'DisperserModel',
           'NIRCAMGrismModel', 'NIRISSGrismModel', 'WaveCorrModel']


class _SimpleModel(ReferenceFileModel):
    """
    A model for a reference file of type "distortion".
    """
    schema_url = None
    reftype = None

    def __init__(self, init=None, model=None, input_units=None, output_units=None, **kwargs):

        super(_SimpleModel, self).__init__(init=init, **kwargs)
        if model is not None:
            self.model = model
        if input_units is not None:
            self.meta.input_units = input_units
        if output_units is not None:
            self.meta.output_units = output_units
        if init is None:
            try:
                self.populate_meta()
            except NotImplementedError:
                pass

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        """
        Subclasses can overwrite this to populate specific meta keywords.
        """
        raise NotImplementedError

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.model, Model) or all([isinstance(m, Model) for m in self.model])
            assert self.meta.instrument.name in ["NIRCAM", "NIRSPEC", "MIRI", "TFI", "FGS", "NIRISS"]
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class DistortionModel(_SimpleModel):
    """
    A model for a reference file of type "distortion".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/distortion.schema"
    reftype = "distortion"

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.meta.input_units, (str, u.NamedUnit))
            assert isinstance(self.meta.output_units, (str, u.NamedUnit))
            if self.meta.instrument.name == 'NIRCAM':
                assert self.meta.instrument.module is not None
                assert self.meta.instrument.channel is not None
                assert self.meta.instrument.p_pupil is not None
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class DistortionMRSModel(ReferenceFileModel):
    """
    A model for a reference file of type "distortion" for the MIRI MRS.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/distortion_mrs.schema"
    reftype = "distortion"

    def __init__(self, init=None, x_model=None, y_model=None, alpha_model=None, beta_model=None,
                 bzero=None, bdel=None, input_units=None, output_units=None, **kwargs):

        super().__init__(init=init, **kwargs)

        if x_model is not None:
            self.x_model = x_model
        if y_model is not None:
            self.y_model = y_model
        if alpha_model is not None:
            self.alpha_model = alpha_model
        if beta_model is not None:
            self.beta_model = beta_model
        if bzero is not None:
            self.bzero = bzero
        if bdel is not None:
            self.bdel = bdel
        if input_units is not None:
            self.meta.input_units = input_units
        if output_units is not None:
            self.meta.output_units = output_units
        if init is None:
            try:
                self.populate_meta()
            except NotImplementedError:
                pass

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "MIRI"
        self.meta.exposure.type = "MIR_MRS"
        self.meta.input_units = u.pix
        self.meta.output_units = u.arcsec

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.meta.input_units, (str, u.NamedUnit))
            assert isinstance(self.meta.output_units, (str, u.NamedUnit))
            assert self.meta.instrument.name == "MIRI"
            assert self.meta.exposure.type == "MIR_MRS"
            assert self.meta.instrument.channel in ("12", "34", "1", "2", "3", "4")
            assert self.meta.instrument.band in ("SHORT", "LONG", "MEDIUM", "SHORT-LONG", "SHORT-MEDIUM",
                                                 "MEDIUM-SHORT", "MEDIUM-LONG", "LONG-SHORT", "LONG-MEDIUM")
            assert self.meta.instrument.detector in ("MIRIFUSHORT", "MIRIFULONG")
            assert all([isinstance(m, Model) for m in self.x_model])
            assert all([isinstance(m, Model) for m in self.y_model])
            assert all([isinstance(m, Model) for m in self.alpha_model])
            assert all([isinstance(m, Model) for m in self.beta_model])
            assert len(self.abv2v3_model.model) == 2
            assert len(self.abv2v3_model.channel_band) == 2
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class SpecwcsModel(_SimpleModel):
    """
    A model for a reference file of type "specwcs".

    Notes
    -----
    For NIRISS and NIRCAM WFSS modes the specwcs file is
    used during extract_2D. See NIRCAMGrismModel and
    NIRISSGrismModel.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/specwcs.schema"
    reftype = "specwcs"

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.meta.input_units, (str, u.NamedUnit))
            assert isinstance(self.meta.output_units, (str, u.NamedUnit))
            assert self.meta.instrument.name in ["NIRCAM",
                                                 "NIRSPEC",
                                                 "MIRI",
                                                 "TFI",
                                                 "FGS",
                                                 "NIRISS"]
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class NIRCAMGrismModel(ReferenceFileModel):
    """
    A model for a reference file of type "specwcs" for NIRCAM WFSS.

    This reference file contains the models for wave, x, and y polynomial
    solutions that describe dispersion through the grism.

    Parameters
    ----------
    displ: `~astropy.modeling.Model`
          Nircam Grism wavelength dispersion model
    dispx : `~astropy.modeling.Model`
          Nircam Grism row dispersion model
    dispy : `~astropy.modeling.Model`
          Nircam Grism column dispersion model
    invdispl : `~astropy.modeling.Model`
          Nircam Grism inverse wavelength dispersion model
    invdispx : `~astropy.modeling.Model`
          Nircam Grism inverse row dispersion model
    invdispy : `~astropy.modeling.Model`
          Nircam Grism inverse column dispersion model
    orders : `~astropy.modeling.Model`
          NIRCAM Grism orders, matched to the array locations of the
          dispersion models

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/specwcs_nircam_grism.schema"
    reftype = "specwcs"

    def __init__(self, init=None,
                 displ=None,
                 dispx=None,
                 dispy=None,
                 invdispl=None,
                 invdispx=None,
                 invdispy=None,
                 orders=None,
                 **kwargs):
        super().__init__(init=init, **kwargs)

        if init is None:
            self.populate_meta()
        if displ is not None:
            self.displ = displ
        if dispx is not None:
            self.dispx = dispx
        if dispy is not None:
            self.dispy = dispy
        if invdispl is not None:
            self.invdispl = invdispl
        if invdispx is not None:
            self.invdispx = invdispx
        if invdispy is not None:
            self.invdispy = invdispy
        if orders is not None:
            self.orders = orders

    def populate_meta(self):
        self.meta.instrument.name = "NIRCAM"
        self.meta.exposure.type = "NRC_WFSS"
        self.meta.reftype = self.reftype

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.meta.input_units, (str, u.NamedUnit))
            assert isinstance(self.meta.output_units, (str, u.NamedUnit))
            assert self.meta.instrument.name == "NIRCAM"
            assert self.meta.exposure.type == "NRC_WFSS"
            assert self.meta.reftype == self.reftype
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")


class NIRISSGrismModel(ReferenceFileModel):
    """
    A model for a reference file of type "specwcs" for NIRISS grisms.

    Parameters
    ----------
    displ: `~astropy.modeling.Model`
          NIRISS Grism wavelength dispersion model
    dispx : `~astropy.modeling.Model`
          NIRISS Grism row dispersion model
    dispy : `~astropy.modeling.Model`
          NIRISS Grism column dispersion model
    invdispl : `~astropy.modeling.Model`
          NIRISS Grism inverse wavelength dispersion model
    invdispx : `~astropy.modeling.Model`
          NIRISS Grism inverse row dispersion model
    invdispy : `~astropy.modeling.Model`
          NIRISS Grism inverse column dispersion model
    orders : `~astropy.modeling.Model`
          NIRISS Grism orders, matched to the array locations of the
          dispersion models
    fwcpos_ref : float
        The reference value for the filter wheel position
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/specwcs_niriss_grism.schema"
    reftype = "specwcs"

    def __init__(self, init=None,
                 displ=None,
                 dispx=None,
                 dispy=None,
                 invdispl=None,
                 orders=None,
                 fwcpos_ref=None,
                 **kwargs):
        super().__init__(init=init, **kwargs)

        if init is None:
            self.populate_meta()
        if displ is not None:
            self.displ = displ
        if dispx is not None:
            self.dispx = dispx
        if dispy is not None:
            self.dispy = dispy
        if invdispl is not None:
            self.invdispl = invdispl
        if orders is not None:
            self.orders = orders
        if fwcpos_ref is not None:
            self.fwcpos_ref = fwcpos_ref

    def populate_meta(self):
        self.meta.instrument.name = "NIRISS"
        self.meta.instrument.detector = "NIS"
        self.meta.exposure.type = "NIS_WFSS"
        self.meta.reftype = self.reftype

    def validate(self):
        super(NIRISSGrismModel, self).validate()
        try:
            assert isinstance(self.meta.input_units, (str, u.NamedUnit))
            assert isinstance(self.meta.output_units, (str, u.NamedUnit))
            assert self.meta.instrument.name == "NIRISS"
            assert self.meta.exposure.type == "NIS_WFSS"
            assert self.meta.instrument.detector == "NIS"
            assert self.meta.reftype == self.reftype
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")


class RegionsModel(ReferenceFileModel):
    """
    A model for a reference file of type "regions".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/regions.schema"
    reftype = "regions"

    def __init__(self, init=None, regions=None, **kwargs):
        super().__init__(init=init, **kwargs)
        if regions is not None:
            self.regions = regions
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "MIRI"
        self.meta.exposure.type = "MIR_MRS"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.regions, (np.ndarray, NDArrayType))
            assert self.meta.instrument.name == "MIRI"
            assert self.meta.exposure.type == "MIR_MRS"
            assert self.meta.instrument.channel in ("12", "34", "1", "2", "3", "4")
            assert self.meta.instrument.band in ("SHORT", "MEDIUM", "LONG", "SHORT-LONG", "SHORT-MEDIUM",
                                                 "MEDIUM-SHORT", "MEDIUM-LONG", "LONG-SHORT", "LONG-MEDIUM")
            assert self.meta.instrument.detector in ("MIRIFUSHORT", "MIRIFULONG")
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class WavelengthrangeModel(ReferenceFileModel):
    """
    A model for a reference file of type "wavelengthrange".

    The model is used by MIRI, NIRSPEC, NIRCAM, and NIRISS.


    Parameters
    ----------
    wrange : list
        Contains a list of [order, filter, min wave, max wave]
    order : list
        A list of orders that are available and described in the file
    extract_orders : list
        A list of filters and the orders that should be extracted by default
    wunits : `~astropy.units`
        The units for the wavelength data

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/wavelengthrange.schema"
    reftype = "wavelengthrange"

    def __init__(self, init=None, wrange_selector=None, wrange=None,
                 order=None, extract_orders=None, wunits=None, **kwargs):

        super().__init__(init=init, **kwargs)
        if wrange_selector is not None:
            self.waverange_selector = wrange_selector
        if wrange is not None:
            self.wavelengthrange = wrange
        if order is not None:
            self.order = order
        if extract_orders is not None:
            self.extract_orders = extract_orders
        if wunits is not None:
            self.meta.wavelength_units = wunits

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file")

    def validate(self):
        super().validate()
        try:
            assert self.meta.instrument.name in ("MIRI", "NIRSPEC", "NIRCAM", "NIRISS")
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)

    def get_wfss_wavelength_range(self, filter, orders):
        """ Retrieve the wavelength range for a WFSS observation.

        Parameters
        ----------
        filter : str
            Filter for which to retrieve the wavelength range.
        orders : list
            List of spectral orders

        Returns
        -------
        wave_range : dict
            Pairs of {order: (wave_min, wave_max)} for each order and the specific filter.
        """
        wave_range = {}
        for order in orders:
            range_select = [(x[2], x[3]) for x in self.wavelengthrange if (x[0] == order and x[1] == filter)]
            wave_range[order] = range_select[0]
        return wave_range


class FPAModel(ReferenceFileModel):
    """
    A model for a NIRSPEC reference file of type "fpa".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/fpa.schema"
    reftype = "fpa"

    def __init__(self, init=None, nrs1_model=None, nrs2_model=None, **kwargs):

        super().__init__(init=init, **kwargs)
        if nrs1_model is not None:
            self.nrs1_model = nrs1_model
        if nrs2_model is not None:
            self.nrs2_model = nrs2_model
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()
        try:
            assert isinstance(self.nrs1_model, Model)
            assert isinstance(self.nrs2_model, Model)
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class IFUPostModel(ReferenceFileModel):
    """
    A model for a NIRSPEC reference file of type "ifupost".

    Parameters
    ----------
    init : str
        A file name.
    slice_models : dict
        A dictionary with slice transforms with the following entries
        {"slice_N": {'linear': astropy.modeling.Model,
        ...         'xpoly': astropy.modeling.Model,
        ...         'xpoly_distortion': astropy.modeling.Model,
        ...         'ypoly': astropy.modeling.Model,
        ...         'ypoly_distortion': astropy.modeling.Model,
        ...         }}

    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ifupost.schema"
    reftype = "ifupost"

    def __init__(self, init=None, slice_models=None, **kwargs):

        super().__init__(init=init, **kwargs)
        if slice_models is not None:
            if len(slice_models) != 30:
                raise ValueError("Expected 30 slice models, got {0}".format(len(slice_models)))
            else:
                for key, val in slice_models.items():
                    setattr(self, key, val)
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.type = "NRS_IFU"
        self.meta.exposure.p_exptype = "NRS_IFU"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()


class IFUSlicerModel(ReferenceFileModel):
    """
    A model for a NIRSPEC reference file of type "ifuslicer".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ifuslicer.schema"
    reftype = "ifuslicer"

    def __init__(self, init=None, model=None, data=None, **kwargs):

        super().__init__(init=init, **kwargs)
        if model is not None:
            self.model = model
        if data is not None:
            self.data = data
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.type = "NRS_IFU"
        self.meta.exposure.p_exptype = "NRS_IFU"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()


class MSAModel(ReferenceFileModel):
    """
    A model for a NIRSPEC reference file of type "msa".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/msa.schema"
    reftype = "msa"

    def __init__(self, init=None, models=None, data=None, **kwargs):
        super().__init__(init=init, **kwargs)
        if models is not None and data is not None:
            self.Q1 = {'model': models['Q1'], 'data': data['Q1']}
            self.Q2 = {'model': models['Q2'], 'data': data['Q2']}
            self.Q3 = {'model': models['Q3'], 'data': data['Q3']}
            self.Q4 = {'model': models['Q4'], 'data': data['Q4']}
            self.Q5 = {'model': models['Q5'], 'data': data['Q5']}
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()


class DisperserModel(ReferenceFileModel):
    """
    A model for a NIRSPEC reference file of type "disperser".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/disperser.schema"
    reftype = "disperser"

    def __init__(self, init=None, angle=None, gwa_tiltx=None, gwa_tilty=None,
                 kcoef=None, lcoef=None, tcoef=None, pref=None, tref=None,
                 theta_x=None, theta_y=None, theta_z=None,
                 groovedensity=None, **kwargs):
        super().__init__(init=init, **kwargs)
        if groovedensity is not None:
            self.groovedensity = groovedensity
        if angle is not None:
            self.angle = angle
        if gwa_tiltx is not None:
            self.gwa_tiltx = gwa_tiltx
        if gwa_tilty is not None:
            self.gwa_tilty = gwa_tilty
        if kcoef is not None:
            self.kcoef = kcoef
        if lcoef is not None:
            self.lcoef = lcoef
        if tcoef is not None:
            self.tcoef = tcoef
        if pref is not None:
            self.pref = pref
        if tref is not None:
            self.tref = tref
        if theta_x is not None:
            self.theta_x = theta_x
        if theta_y is not None:
            self.theta_y = theta_y
        if theta_z is not None:
            self.theta_z = theta_z
        if gwa_tiltx is not None:
            self.gwa_tiltx = gwa_tiltx
        if gwa_tilty is not None:
            self.gwa_tilty = gwa_tilty
        if init is None:
            self.populate_meta()

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"

    def to_fits(self):
        raise NotImplementedError("FITS format is not supported for this file.")

    def validate(self):
        super().validate()
        try:
            assert self.meta.instrument.grating in ["G140H", "G140M", "G235H", "G235M",
                                                    "G395H", "G395M", "MIRROR", "PRISM"]
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class FilteroffsetModel(ReferenceFileModel):
    """
    A model for filter-dependent boresight offsets.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/filteroffset.schema"
    reftype = "filteroffset"

    def __init__(self, init=None, filters=None, instrument=None, **kwargs):
        super().__init__(init, **kwargs)
        if filters is not None:
            self.filters = filters
            if instrument is None or instrument not in ("NIRCAM", "MIRI", "NIRISS"):
                raise ValueError("Please specify the instrument")
            self.meta.instrument.name = instrument

    def populate_meta(self):
        self.meta.reftype = self.reftype
        if self.meta.instrument.name == "MIRI":
            self.meta.instrument.detector = "MIRIMAGE"
            self.meta.instrument.pfilter = "F1130W|F1140C|F2300C|F2100W|F1800W|\
            F1550C|F560W|F2550WR|FND|F2550W|F1500W|F1000W|F1065C|F770W|F1280W|"
        elif self.meta.instrument.name == "NIRCAM":
            self.meta.instrument.pfilter = "F070W|F090W|F115W|F140M|F150W|\
            F162M|F164N|F150W2|F182M|F187N|F200W|F210M|F212N|"
        elif self.meta.instrument.name == "NIRISS":
            self.meta.instrument.pfilter = "F070W|F115W|F140M|F150W|F158M|\
            F200W|F277W|F356W|F380M|F430M|F444W|F480M|"
        else:
            raise ValueError(f"Unsupported instrument: {self.meta.instrument.name}")

    def validate(self):
        super().validate()

        instrument_name = self.meta.instrument.name
        nircam_channels = ["SHORT", "LONG"]
        nircam_module = ["A", "B"]
        if instrument_name not in ['MIRI', 'NIRCAM', 'NIRISS']:
            self.print_err('Expected "meta.instrument.name" to be one of "NIRCAM, "MIRI" or "NIRISS"')
        if instrument_name == "MIRI" and self.meta.instrument.detector != "MIRIMAGE":
            self.print_err('Expected detector to be MIRIMAGE for instrument MIRI')
        elif instrument_name == "NIRCAM":
            if self.meta.instrument.channel not in nircam_channels:
                self.print_err(f'Expected meta.instrument.channel for instrument NIRCAM to be one of {nircam_channels}')
            if self.meta.instrument.module not in nircam_module:
                self.print_err(f'Expected meta.instrument.module for instrument NIRCAM to be one of {nircam_module}')


class IFUFOREModel(_SimpleModel):
    """
    A model for a NIRSPEC reference file of type "ifufore".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ifufore.schema"
    reftype = "ifufore"

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.type = "NRS_IFU"
        self.meta.exposure.p_exptype = "NRS_IFU"


class CameraModel(_SimpleModel):
    """
    A model for a reference file of type "camera".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/camera.schema"
    reftype = 'camera'

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"


class CollimatorModel(_SimpleModel):
    """
    A model for a reference file of type "collimator".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/collimator.schema"
    reftype = 'collimator'

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"


class OTEModel(_SimpleModel):
    """
    A model for a reference file of type "ote".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ote.schema"
    reftype = 'ote'

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"


class FOREModel(_SimpleModel):
    """
    A model for a reference file of type "fore".
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/fore.schema"
    reftype = 'fore'

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"
        self.meta.instrument.p_detector = "NRS1|NRS2|"
        self.meta.exposure.p_exptype = "NRS_TACQ|NRS_TASLIT|NRS_TACONFIRM|\
        NRS_CONFIRM|NRS_FIXEDSLIT|NRS_IFU|NRS_MSASPEC|NRS_IMAGE|NRS_FOCUS|\
        NRS_MIMF|NRS_MSATA|NRS_WATA|NRS_LAMP|NRS_BRIGHTOBJ|"
        self.meta.exposure.type = "N/A"

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def validate(self):
        super().validate()
        try:
            assert self.meta.instrument.filter in ["CLEAR", "F070LP", "F100LP", "F110W",
                                                   "F140X", "F170LP", "F290LP"]
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)


class WaveCorrModel(ReferenceFileModel):

    reftype = "wavecorr"
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/wavecorr.schema"

    def __init__(self, init=None, apertures=None, **kwargs):
        super().__init__(init, **kwargs)
        if apertures is not None:
            self.apertures = apertures
        if init is None:
            self.populate_meta()

    @property
    def aperture_names(self):
        return [getattr(ap, 'aperture_name') for ap in self.apertures]

    def populate_meta(self):
        self.meta.instrument.name = "NIRSPEC"

    def on_save(self, path=None):
        self.meta.reftype = self.reftype

    def validate(self):
        super().validate()
        try:
            assert self.aperture_names is not None
            assert self.apertures is not None
        except AssertionError:
            if self._strict_validation:
                raise
            else:
                warnings.warn(traceback.format_exc(), ValidationWarning)



# The schema editor is designed to be run as a command line script. It can be
# run either interactively or not. To run it non-interactively, you must set
# the options when creating  a new editor object, as in the following
# example. The options are listed in the __doc__ comment for the Schema_editor
# class.
#
# from jwst.datamodels import schema_editor
# editor = schema_editor.Schema_editor(add=True, delete=True,
#                                      edit=True, rename=True)
# editor.change()
#
# To list the changes that would be made instead of making the changes,
# add the the and log options. The changes will be written to the file
# passed as the log option.
#
# from jwst.datamodels import schema_editor
# editor = schema_editor.Schema_editor(add=True, delete=True,
#                                      edit=True, rename=True,
#                                      list=True, log="changes.txt")
# editor.change()
#
# To run it interactively, first create an options object and pass it as an
# argument to the change method:
#
# from jwst.datamodels import schema_editor
# options = schema_editor.Options()
# editor = schema_editor.Schema_editor()
# editor.change(options)
#
# You can also save the options selected in a previous run to a file.
# Previous values are provided as defaults and schema changes that were
# previously again are not queried for again. The options file is written
# to your home directory unless the filename is absolute. The filename is
# passed to the options object when initialized:
#
# from jwst.datamodels import schema_editor
# options = schema_editor.Options(filename=".schema_editor_options")
# editor = schema_editor.Schema_editor()
# editor.change(options)
#
# It is possible to set all the options to the schema editor on the command
# line. Options that are not set are assumed to be false. To list the
# options, invoke the script with --help on the command line.

import argparse
from collections import OrderedDict
import datetime
import inspect
import os
import os.path
import re
import sys
from urllib.parse import urlparse
import yaml

from asdf import schema as aschema
from asdf import generic_io
from asdf import reference
from asdf import treeutil

from . import JwstDataModel


__all__ = ["enquote", "is_long_line", "leading_length", "new_path",
           "save_complex_list", "save_dictionary", "save_long_line",
           "save_scalar", "save_short_line", "save_simple_list",
           "Keyword_db", "Model_db", "Options", "Schema_editor",]


def enquote(value):
    """
    Put quote marks around value if it contains special characters
    """
    if re.search(r'[^a-zA-Z0-9_ \t\v\(\)]', value):
        value = re.sub("'", "''", value)
        value = "'" + value + "'"
    return value


def is_long_line(prefix, value, sep, max_length=80):
    """
    Test if a line is so long it needs to be folded onto the next,
    """
    value = str(value)
    if value.find(sep) >= 0:
        long_line = prefix + ": " + value
        return len(long_line) > max_length
    else:
        return False


def leading_length(prefix):
    return len(prefix) - len(prefix.lstrip())


def new_path(path, model_name):
    """
    Add  the model name to the path if it is not None
    """
    new_path = path[:]
    if model_name is not None:
        new_path.append(model_name)
    return new_path


def save_complex_list(fd, name, values, leading):
    fd.write(leading + name + ":\n")
    for value in values:
        save_dictionary(fd, value, leading=leading + '- ')


def save_dictionary(fd, schema, leading=''):
    delayed = []
    for (name, value) in schema.items():
        if name == "$schema":
            delayed.append(name)

        elif name == "properties" or isinstance(value, dict):
            fd.write(leading + name + ":\n")
            save_dictionary(fd, value, leading=leading + '  ')

        elif isinstance(value, list):
            if len(value) > 0 and isinstance(value[0], dict):
                save_complex_list(fd, name, value, leading)
            else:
                prefix = leading + name + ": "
                fd.write(save_simple_list(prefix, value))

        else:
            prefix = leading + name + ": "
            if name in ("title", "description"):
                value = enquote(value)
            fd.write(save_scalar(prefix, value))

        leading = leading.replace('-', ' ')

    for name in delayed:
        value = schema[name]
        if isinstance(value, list):
            fd.write(save_simple_list(leading, name, value))
        else:
            prefix = leading + name + ": "
            fd.write(save_scalar(prefix, value))


def save_long_line(prefix, value, sep1, sep2, max_length=80):
    """
    Converta long line into a folded string
    """
    long_line = prefix
    values = value.split(sep1)

    start = True
    line_start = 0
    leading = ' ' * (leading_length(prefix) + 2)

    for value in values:
        nl = value.find('\n')
        if nl >= 0:
            long_line += value[0:nl + 1]
            value = value[nl + 1:]
            line_start = len(long_line)

        if start:
            long_line += value
            start = False
        else:
            line_length = (len(long_line) + len(value) + len(sep1)
                           - line_start)

            if line_length < max_length:
                long_line += sep1 + value
            else:
                line_start = len(long_line) + len(sep2)
                long_line += sep2 + leading + value

    long_line += '\n'
    return long_line


def save_scalar(prefix, value):
    """
    Convert a scalar to a possibly folded string
    """
    sep1 = '|'
    sep2 = sep1 + '\\\n'
    if is_long_line(prefix, value, sep1):
        value = '"' + str(value) + '"'
        line = save_long_line(prefix, value, sep1, sep2)
    else:
        line = save_short_line(prefix, value)
    return line


def save_short_line(prefix, value):
    short_line = prefix + str(value) + '\n'
    return short_line


def save_simple_list(prefix, values, max_length=80):
    """
    Convert a list, possibly folding at the separator
    """
    sep1 = ', '
    sep2 = sep1 + '\n'
    vstr = [str(v) for v in values]
    value = '[' + sep1.join(vstr) + ']'
    if is_long_line(prefix, value, sep1):
        line = save_long_line(prefix, value, sep1, sep2)
    else:
        line = save_short_line(prefix, value)
    return line


class Keyword_db:
    def __init__(self, directory=""):
        """
        Combine all the top dbs in a directory into a single db

        Parameters
        ----------
        directory: The directory containing the downloaded schema files.
                   If blank, use the current directory
        """

        directory = self.find_directory(directory)
        if directory is None:
            raise ValueError("Cannot locate keyword database directory")

        self.schema = None
        for filename in os.listdir(directory):
            if filename.startswith("top."):
                keyword_db = os.path.abspath(os.path.join(directory, filename))
                schema = aschema.load_schema(keyword_db, resolve_references=False)
                if self.schema is None:
                    self.schema = schema
                else:
                    try:
                        self.combine_schemas(schema)
                    except ValueError as err:
                        raise ValueError(filename + ": " + str(err))

        self.resolve_references(directory)

    def combine_schemas(self, other_schema):
        """
        Combine another schema into the keyword database schema
        """
        def combine_dictionaries(this_schema, other_schema):
            error_msg = "Unrecognized field in schema: "

            for (name, other_subschema) in other_schema.items():
                if "properties" not in other_subschema:
                    raise ValueError(error_msg + name)

                if name in this_schema:
                    this_subschema = this_schema[name]

                    if "properties" not in this_subschema:
                        raise ValueError(error_msg + name)

                    if ("allOf" in this_subschema["properties"] or
                            "allOf" in other_subschema["properties"]):
                        try:
                            combine_lists(this_subschema, other_subschema)
                        except ValueError:
                            raise ValueError(error_msg + name)

                    elif ("$ref" in this_subschema["properties"] and
                            "$ref" in other_subschema["properties"]):
                        if (this_subschema["properties"]["$ref"] !=
                                other_subschema["properties"]["$ref"]):
                            try:
                                combine_lists(this_subschema, other_subschema)
                            except ValueError:
                                raise ValueError(error_msg + name)

                    else:
                        raise ValueError(error_msg + name)

                else:
                    this_schema[name] = other_subschema

        def combine_lists(this_schema, other_schema):
            combination = {}
            for schema in (this_schema, other_schema):
                if "allOf" in schema["properties"]:
                    for item in schema["properties"]["allOf"]:
                        if "$ref" in item:
                            ref = item["$ref"]
                            combination[ref] = item
                        else:
                            raise ValueError

                elif "$ref" in schema["properties"]:
                    ref = schema["properties"]["$ref"]
                    combination[ref] = schema["properties"]

                else:
                    raise ValueError
            this_schema["properties"] = {"allOf": list(combination.values())}

        this_schema = self.schema["properties"]["meta"]["properties"]
        other_schema = other_schema["properties"]["meta"]["properties"]
        combine_dictionaries(this_schema, other_schema)

    def create_dict(self):
        """
        Create a dictionary mapping fits keyword names to keyword db paths
        """
        def recurse(keyword_schema, keyword_dict, path):
            if "properties" in keyword_schema:
                keyword_schema = keyword_schema["properties"]

            for keyword_name, keyword_subschema in keyword_schema.items():
                if isinstance(keyword_subschema, dict):
                    if "fits_hdu" in keyword_subschema:
                        if keyword_subschema["fits_hdu"] in ('PRIMARY', 'SCI'):
                            if "fits_keyword" in keyword_subschema:
                                # Save the path to any dictionary
                                # with a fits_keyword field
                                keyword_fits_name = keyword_subschema["fits_keyword"]
                                keyword_fits_name = keyword_fits_name.strip()
                                keyword_dict[keyword_fits_name] = \
                                    '.'.join(new_path(path, keyword_name))

                    else:
                        # Skip "standard" and "misc",
                        # they are not in the model schema
                        if keyword_name not in ("standard", "misc"):
                            recurse(keyword_subschema,
                                    keyword_dict,
                                    new_path(path, keyword_name))

                elif (isinstance(keyword_subschema, list) and
                        keyword_name == "allOf"):
                    # Need to combine keyword db schemas if they are under
                    # an "allOf"
                    merged_subschema = self.merge_schemas(keyword_subschema)
                    recurse(merged_subschema, keyword_dict, path)

        path = []
        keyword_dict = {}
        keyword_schema = self.schema
        recurse(keyword_schema, keyword_dict, path)

        return keyword_dict

    def find_directory(self, directory):
        """
        Return full path to requested directory

        Notes
        -----
        There was originally some functionality involving a prefix.
        This functionality has been removed.
        """
        if len(directory) == 0:
            directory = os.getcwd()

        chosen_dir = os.path.join(os.path.abspath(directory), "")

        return chosen_dir

    def merge_schemas(self, schema):
        """
        Merge the contents of two schemas, including enums in both
        """
        def merge_dictionaries(merged_subschema, dictionary):
            for name, subdictionary in dictionary.items():
                if name in merged_subschema:
                    if "enum" in merged_subschema[name] and "enum" in subdictionary:
                        merge_enums(merged_subschema[name], subdictionary)
                    else:
                        merged_subschema[name] = subdictionary
                else:
                    merged_subschema[name] = subdictionary

        def merge_enums(merged_subschema, dictionary):
            merged_subschema["enum"] = list(set(merged_subschema["enum"]) |
                                            set(dictionary["enum"]))

        merged_subschema = OrderedDict()
        for dictionary in schema:
            if merged_subschema:
                merge_dictionaries(merged_subschema, dictionary)
            else:
                merged_subschema = dictionary

        return merged_subschema

    def resolve_references(self, url):
        """
        Resolve urls in the schema
        """
        def resolve_refs(node, json_id):
            if json_id is None:
                json_id = url
            if isinstance(node, dict) and '$ref' in node:
                suburl = generic_io.resolve_uri(json_id, node['$ref'])
                parts = urlparse(suburl)
                fragment = parts.fragment
                if len(fragment):
                    suburl_path = suburl[:-(len(fragment) + 1)]
                else:
                    suburl_path = suburl

                try:
                    subschema = aschema.load_schema(suburl_path,
                                                    resolve_references=True)
                except IOError:
                    print("Could not read " + suburl_path)
                    subschema = OrderedDict()

                subschema_fragment = reference.resolve_fragment(
                    subschema, fragment)

                return subschema_fragment
            return node

        self.schema = treeutil.walk_and_modify(self.schema, resolve_refs)


class Model_db:
    """
    Load the list of datamodels schema files from the schema directory

    Parameters
    ----------
    exclude: None or [str[,...]]
        List of `DataModel` schemas to ignore.

    Attributes
    ----------
    base_url: str
        The location of the `DataModel` schemas

    schema_files: [str[...]]
        The schema files loaded.
    """

    def __init__(self, exclude=None):
        if exclude is None:
            exclude = []

        source_file = os.path.abspath(inspect.getfile(JwstDataModel))
        self.base_url = os.path.join(os.path.dirname(source_file),
                                     'schemas', '')
        self.schema_files = []

        for filename in os.listdir(self.base_url):
            if filename.endswith(".yaml") and filename not in exclude:
                self.schema_files.append(filename)

    def __len__(self):
        return len(self.schema_files)

    def __iter__(self):
        """
        Return an iterator to the list of schema files
        """
        return iter(self.schema_files)

    def read(self, schema_file):
        """
        Read a schema file into memory
        """
        fname = os.path.join(self.base_url, schema_file)
        schema = aschema.load_schema(fname, resolve_references=False)

        return schema

    def save(self, schema_file, schema):
        """
        Save the modified schema back to disk
        """
        with open(schema_file, mode='w') as fd:
            save_dictionary(fd, schema)


class Options:
    """
    Get options for running the schema editor from file, command line arguments,
    and interactive user input.
    """

    preamble = \
        """
         Welcome to the schema editor. The editor changes datamodel schema
         files to match information in the keyword database. You have control
         over which changes are made when the editor is run.

         Before starting, refer to the process document found at:
         https://innerspace.stsci.edu/display/SCSB/Keyword+Dictionary+JSON+vs.+YAML+Comparison+Process


         Retrieve the JWST Keyword Database as described. Then run this script.
         First it will ask you for the name of the directory containing the
         keyword database and the output directory you want the changed model
         schemas written to. Then it will ask you what kind of changes you wish
         to make to the schema files. (Additions to the schema, deletions, and
         so on.) Then it will determine the differences, display them one at a
         time, and ask if you want to make the change. If you say yes, the
         change will be made to the schema. Finally, it will create a new
         subdirectory in the output directory whose name starts with "schemas"
         and write all modified schemas to that directory.

         If schema are modified, and keywords are omitted, a file
         `aaa_omitted_keywords.yaml` will be created. This file can be used as
         the input to the `--omit_file` option

         The first two inputs are the names of the input and output directories.
         If you leave them blank, this script will use the current directory.
         All the following input to the script should either be y (yes) or n (no).
         Each question has a default answer which will be displayed as a
         capital letter (Y or N). If you hit return, the script will use
         the default value.

        """

    run_script = "do you want to continue running the editor"

    # Command line options. 2 or 3-tuple where:
    #    - name: Parameter name and full option name
    #    - help: The help text
    #    - Create a short option. If not present or True, a single character option,
    #      based on the first character of the name, is created.
    prompts = (
        ("input", "directory name containing keyword database"),
        ("output", "directory name model schemas will be written to"),
        ("add", "fields that exist in the keyword db but are not in the model schemas"),
        ("delete", "fields in the model but not in the keyword db from the model"),
        ("edit", "fields found in both to match the values in the keyword db"),
        ("rename", "fields in the model to match the names in the keyword db"),
        ("list", "changes without making them"),
        ("query", "for approval of changes before they are made"),
        ("omit_file", "containing the list of parameters to omit", False),
        ("exclude_file", "containing the list of DataModel schema to exclude from the comparison", False)
    )

    def __init__(self, filename=None):
        """
        Create option file reader / writer

        Parameters
        ----------
        filename: The name of the file string options from the previous run
                  If set to None, no file will be read or written.
        """
        self.batch = len(sys.argv) > 1

        if filename is None:
            self.filename = filename
            self.first_time = True

        else:
            if os.path.isabs(filename):
                self.filename = filename
            else:
                home = os.path.expanduser("~")
                self.filename = os.path.join(home, filename)
            self.first_time = not os.path.isfile(self.filename)

    def coerce_type(self, value, current_value):
        """
        Coerce value to type of the current value of a parameter
        """
        if isinstance(current_value, bool):
            if isinstance(value, str):
                value = value[0]
                value = value.lower()
                if value == "y":
                    value = True
                elif value == "n":
                    value = False
                else:
                    value = None

            elif not isinstance(value, bool):
                value = None

        elif isinstance(current_value, set):
            if hasattr(value, "__iter__"):
                value = set(value)
            else:
                value = set((value, ))

        return value

    def get(self, editor):
        """
        Initialize fields from defaults, command line, and options file

        Parameters
        ----------
        editor: The schema editor whose options are being set
        """
        parameters = self.get_parameters(editor)
        if self.filename is not None and not self.first_time:
            parameters = self.read(parameters)

        if self.batch:
            parameters = self.parse_command_line(parameters)
        else:
            if self.first_time:
                if not self.query_run(self.preamble, self.run_script):
                    return False

            parameters = self.query_options(parameters)

        self.set_parameters(editor, parameters)
        return True

    def get_parameters(self, editor):
        """
        Save the object's fields into a dictionary
        """
        parameters = {}
        for name, prompt, *abbrev in self.prompts:
            parameters[name] = getattr(editor, name)

        return parameters

    def parse_command_line(self, parameters):
        """
        Parse the parameters from the command line
        """
        parser = argparse.ArgumentParser()

        for prompt in self.prompts:
            if prompt[1]:
                name = prompt[0]
                help_text = " ".join(prompt[:2])
                full = "--" + name
                if len(prompt) > 2 and not prompt[2]:
                    option_names = (full,)
                else:
                    abbrev = full[1:3]
                    option_names = (abbrev, full)
                if isinstance(parameters[name], bool):
                    parser.add_argument(*option_names, help=help_text,
                                        action="store_true")
                else:
                    parser.add_argument(*option_names, help=help_text)

        args = parser.parse_args()

        for prompt in self.prompts:
            if prompt[1]:
                name = prompt[0]
                try:
                    value = getattr(args, name)
                except AttributeError:
                    value = None

                if value is not None:
                    value = self.coerce_type(value, parameters[name])
                    self.type_check(parameters, name, value)
                    parameters[name] = value

        return parameters

    def parse_parameter(self, parameters, n, name, value):
        """
        Parse a single parameter value, check it, and save it
        """
        if len(name) == 0:
            if n == 0:
                return
            else:
                raise ValueError("Error at line {} in option file".format(n + 1))

        value = value.replace("\t", " ")
        if value.find(" ") >= 0:
            value = value.split()

        value = self.coerce_type(value, parameters[name])
        self.type_check(parameters, name, value)
        parameters[name] = value

        return

    def query_options(self, parameters):
        """
        Query the user for global option values
        """
        for prompt in self.prompts:
            if prompt[1]:
                default_choice = parameters[prompt[0]]
                choice = self.query_user(" ".join(prompt[:2]), default_choice)
                self.type_check(parameters, prompt[0], choice)
                parameters[prompt[0]] = choice
        return parameters

    def query_run(self, preamble, prompt):
        """
        Explain how to run the script and ask the user to continue
        """
        lines = preamble.split("\n")
        for line in lines:
            print(line.lstrip())

        choice = self.query_user(prompt, default_choice=False)
        return choice

    def query_user(self, prompt, default_choice=None):
        """
        Build the prompt and query the user
        """
        if default_choice is True:
            choices = " (Y|n)? "
        elif default_choice is False:
            choices = " (y|N)? "
        elif default_choice is None:
            choices = "? "
        elif len(default_choice) > 0:
            choices = " (%s)? " % default_choice
        else:
            choices = "? "

        choice = None
        prompt = prompt + choices
        while choice is None:
            choice = input(prompt)
            if len(choice) == 0:
                choice = default_choice
            else:
                choice = self.coerce_type(choice, default_choice)

        return choice

    def read(self, parameters):
        """
        Read the contents of an option file into a hash
        """
        n = 0
        name = ""
        value = ""
        with open(self.filename, "r") as fd:
            for line in fd:
                line = line.rstrip()
                if len(line) == 0:
                    name = ""
                    value = ""
                else:
                    if line.startswith(" ") or line.startswith("\t"):
                        value += line
                    else:
                        self.parse_parameter(parameters, n, name, value)

                        (name, eq, value) = line.partition("=")
                        name = name.strip()
                        value = value.strip()
                n += 1

            self.parse_parameter(parameters, n, name, value)

        return parameters

    def set_parameters(self, editor, parameters):
        """
        Set a field on the object after validating it
        """
        for name, value in parameters.items():
            if not hasattr(editor, name):
                invalid_msg = "{0} is not a valid parameter"
                raise ValueError(invalid_msg.format(name))

            value = self.coerce_type(value, parameters[name])
            self.type_check(parameters, name, value)
            setattr(editor, name, value)

    def type_check(self, parameters, name, value):
        """
        Check that the type of a new value agrees with the current type
        """
        badtype = "Invalid parameter value for %s (%s)"
        if isinstance(parameters[name], str):
            if not isinstance(value, str):
                raise ValueError(badtype % (name, value))

        elif isinstance(parameters[name], set):
            if isinstance(value, bool):
                raise ValueError(badtype % (name, value))

        else:
            ptype = type(parameters[name])
            if not isinstance(value, ptype):
                raise ValueError(badtype % (name, value))

    def write(self, editor):
        """
        Write the contents of a hash as an option file

        Parameters
        ----------
        editor: The schema editor whose options are being saved to file
        """
        if self.filename is not None:
            with open(self.filename, "w") as fd:
                parameters = self.get_parameters(editor)

                for name in sorted(parameters.keys()):
                    value = parameters[name]
                    if isinstance(value, bool):
                        if value:
                            value = "yes"
                        else:
                            value = "no"
                    elif isinstance(value, set):
                        value = " " + "\n\t".join(list(value))

                    fd.write("{0} = {1}\n".format(name, value))


class Schema_editor:
    def __init__(self, **keywds):
        """
        Initialize the editor options

        Parameters
        ----------
        All arguments are keywords

        input  : directory name containing keyword database (string)

        output : directory name model schemas will be written to  (string)

        log    : file containing reported modifications to files

        add    : add fields in the keyword db but not in the model to the model (bool)

        delete : delete fields in the model but not in the keyword db from the model (bool)

        edit   : edit fields found in both to match the values in the keyword db (bool)

        rename : rename fields in the model to match the names in the keyword db (bool)

        list   : list changes without making them (bool)

        query  : query for approval of changes before they are made (bool)

        omit   : omit model schema keywords from editing (set)
        """
        # Default values of editor attributes
        self.query = False
        self.list = False
        self.add = False
        self.delete = False
        self.edit = False
        self.rename = False
        self.input = ""
        self.output = ""
        self.log = ""
        self.omit = set()
        self.omit_file = ''
        self.exclude = []
        self.exclude_file = ''
        self.model_db = None

        # Set attributes from keywds
        for name, value in keywds.items():
            if hasattr(self, name):
                setattr(self, name, value)
            else:
                invalid_msg = "{0} is not a valid keyword"
                raise ValueError(invalid_msg.format(name))

        # Initialize fields with info about current file
        self.current_file_name = None
        self.current_file_changed = False

        # Initialize the built in regex pattern
        _builtin_regexes = [
            '', 'NAXIS[0-9]{0,3}', 'BITPIX', 'XTENSION', 'PCOUNT', 'GCOUNT',
            'EXTEND', 'BSCALE', 'BUNIT', 'BZERO', 'BLANK', 'DATAMAX', 'DATAMIN',
            'EXTNAME', 'EXTVER', 'EXTLEVEL', 'GROUPS', 'PTYPE[0-9]',
            'PSCAL[0-9]', 'PZERO[0-9]', 'SIMPLE', 'TFIELDS',
            'TBCOL[0-9]{1,3}', 'TFORM[0-9]{1,3}', 'TTYPE[0-9]{1,3}',
            'TUNIT[0-9]{1,3}', 'TSCAL[0-9]{1,3}', 'TZERO[0-9]{1,3}',
            'TNULL[0-9]{1,3}', 'TDISP[0-9]{1,3}', 'HISTORY'
        ]

        self.builtin_regex = re.compile(
            '|'.join('(^{0}$)'.format(x) for x in _builtin_regexes))

    def add_subschema(self, model_schema, keyword_schema,
                      keyword_dict, fits_dict):
        """
        Add subschema from keyword db to model schema
        """
        for keyword_fits_name in keyword_dict:
            if not self.builtin_fits_keyword(keyword_fits_name):
                if keyword_fits_name not in fits_dict:
                    keyword_path = keyword_dict[keyword_fits_name]
                    keyword_path = keyword_path.split('.')

                    path = []
                    add = False
                    keyword_subschema = keyword_schema
                    model_subschema = model_schema
                    for name in keyword_path:
                        if "properties" in keyword_subschema:
                            keyword_subschema = keyword_subschema["properties"]
                        if "properties" in model_subschema:
                            model_subschema = model_subschema["properties"]

                        test = self.get_keyword_value(keyword_subschema, name)
                        if test is None:
                            break
                        else:
                            keyword_subschema = test

                        test = self.get_keyword_value(model_subschema, name)
                        path.append(name)

                        if test is None:
                            add = True
                            break
                        else:
                            model_subschema = test

                    if add:
                        self.schema_add_subschema(fits_dict,
                                                  model_subschema,
                                                  keyword_subschema,
                                                  path)

    def builtin_fits_keyword(self, fits_name):
        """
        Test if a fits keyword is built in
        """
        return self.builtin_regex.match(fits_name) is not None

    def change(self, options=None):
        """
        Change datamodels schema files to match info in the keyword db

        Parameters
        ----------
        options : An options object, used for querying for attribute values
                  interactively. Do not set if using in a program.
        """

        # Initialize the editor object from file and command line
        self.options = options
        if self.options is None:
            self.query = False
        else:
            if not self.options.get(self):
                return

        # If an omit file was specified, add the contents to the omit set.
        if self.omit_file:
            with open(self.omit_file) as fh:
                self.omit.update(yaml.safe_load(fh))

        # If an exclude file was specified, add the list of `DataModel` schema
        # to the exclusion list.
        if self.exclude_file:
            with open(self.exclude_file) as fh:
                self.exclude.extend(yaml.safe_load(fh))

        # Set output file for messages depending on list,
        # so output can be captured to a file
        if self.query or self.log == "":
            self.fd = sys.stdout
        else:
            self.fd = open(self.log, "w")

        # Is anything going to be done?
        if not any((
                self.add, self.delete, self.edit, self.rename,
        )):
            raise RuntimeError(
                'No operation has been requested. Set at least one of:\n'
                '\tadd, delete, edit, or rename'
            )

        # If not listing, get the output directory setup.
        if not self.list:
            output_dir = self.dated_directory(self.output, "schemas")

        # Parse the keyword database files
        keyword_db = Keyword_db(self.input)
        keyword_dict = keyword_db.create_dict()
        keyword_schema = keyword_db.schema

        # Loop over the model schema files, updating them from the keyword db

        fits_dict = {}
        model_db = Model_db(exclude=self.exclude)
        self.model_db = model_db
        for schema_file in model_db:
            model_path = []
            model_schema = model_db.read(schema_file)
            self.match_fits_keywords(schema_file, model_schema, keyword_schema,
                                     keyword_dict, fits_dict, model_path)

        for schema_file in model_db:
            self.current_file_name = schema_file
            self.current_file_changed = False

            model_path = []
            model_schema = model_db.read(schema_file)
            self.edit_schema(model_schema, keyword_schema,
                             keyword_dict, fits_dict, model_path)

            if self.current_file_changed:
                # current_file_changed is set in report_and_query
                schema_file = os.path.join(output_dir, schema_file)
                model_db.save(schema_file, model_schema)

        schema_file = 'core.schema.yaml'
        self.current_file_changed = False
        model_schema = model_db.read(schema_file)
        self.add_subschema(model_schema, keyword_schema,
                           keyword_dict, fits_dict)

        if self.current_file_changed:
            schema_file = os.path.join(output_dir, schema_file)
            model_db.save(schema_file, model_schema)

        # Save the list of omitted parameters
        if not self.list and self.omit:
            omit_path = os.path.join(output_dir, 'aaa_omitted_keywords.yaml')
            with open(omit_path, 'w') as fh:
                yaml.safe_dump(list(self.omit), fh)

        # Write the object attributes back to disk
        if self.options is not None:
            self.options.write(self)

    def check_type(self, model_schema, default_value):
        """
        Check if default value agrees with data type
        """
        valid = True
        model_type = model_schema.get("type")
        if model_type is not None:
            if model_type == "number":
                try:
                    float(default_value)
                except ValueError:
                    valid = False

            elif model_type == "integer":
                try:
                    int(default_value)
                except ValueError:
                    valid = False

        return valid

    def compare_schema_values(self, keyword_value, model_value):
        """
        Compare iwo values for type specific kind of equality
        """
        if isinstance(keyword_value, str) and isinstance(model_value, str):
            result = (keyword_value.lower().strip() ==
                      model_value.lower().strip())

        elif isinstance(keyword_value, float) or isinstance(model_value, float):
            try:
                result = float(keyword_value) == float(model_value)
            except ValueError:
                result = False

        elif isinstance(keyword_value, int) or isinstance(model_value, int):
            try:
                result = int(keyword_value) == int(model_value)
            except ValueError:
                result = False

        elif isinstance(keyword_value, list) and isinstance(model_value, list):
            result = set(keyword_value) == set(model_value)

        else:
            result = keyword_value == model_value

        return result

    def dated_directory(self, directory, prefix):
        """
        Create a dated directory in a specified directory
        """
        d = datetime.date.today()

        for i in range(100):
            subdirectory = ("%s_%04d_%02d_%02d_%02d" %
                            (prefix, d.year, d.month, d.day, i))

            subdirectory = os.path.join(directory, subdirectory, "")
            if not os.path.exists(subdirectory):
                os.makedirs(subdirectory)
                return subdirectory

        return None

    def edit_schema(self, model_schema, keyword_schema,
                    keyword_dict, fits_dict, model_path):
        """
        Edit model schema to match keywords in keyword db
        """
        properties = self.get_keyword_value(model_schema, "properties")
        if properties is not None:
            model_schema = properties

        model_names = list(model_schema.keys())
        for model_name in model_names:
            model_subschema = model_schema[model_name]
            new_model_path = new_path(model_path, model_name)
            if isinstance(model_subschema, dict):
                model_fits_name = self.get_keyword_value(model_subschema,
                                                         "fits_keyword")

                if model_fits_name is not None:
                    if not self.builtin_fits_keyword(model_fits_name):
                        # Special case for reference file keywords
                        p_name = model_fits_name[0:2] == "P_"

                        keyword_fits_name = fits_dict.get(model_fits_name)
                        if keyword_fits_name is None:
                            keyword_path = None
                        else:
                            keyword_path = keyword_dict.get(keyword_fits_name)

                        if keyword_path is None:
                            if not p_name:
                                self.schema_del_value(model_schema,
                                                      model_name,
                                                      new_model_path)

                        else:
                            keyword_subschema = self.find_subschema(keyword_schema,
                                                                    keyword_path)
                            if keyword_subschema is not None:
                                keyword_path = keyword_path.split('.')
                                self.update_schema_fields(keyword_subschema,
                                                          model_subschema,
                                                          new_model_path)

                                if not p_name and model_name != keyword_path[-1]:
                                    self.schema_rename_value(model_schema,
                                                             keyword_path[-1],
                                                             model_name,
                                                             new_model_path)

                                if model_fits_name != keyword_fits_name:
                                    fits_dict[keyword_fits_name] = \
                                        fits_dict[model_fits_name]
                                    del fits_dict[model_fits_name]

                else:
                    fits_hdu = self.get_keyword_value(model_subschema, "fits_hdu")

                    if fits_hdu is None:
                        self.edit_schema(model_subschema,
                                         keyword_schema,
                                         keyword_dict,
                                         fits_dict,
                                         new_model_path)

                        if not self.has_properties(model_subschema):
                            del model_schema[model_name]

            elif isinstance(model_subschema, list) and model_name == "allOf":
                for model_subsubschema in model_subschema:
                    if isinstance(model_subsubschema, dict):
                        self.edit_schema(model_subsubschema,
                                         keyword_schema,
                                         keyword_dict,
                                         fits_dict,
                                         model_path)

    def find_subschema(self, schema, path):
        """
        Trace a path thru the schema. If not found, return partial
        match if partial is True, None if partial is false
        """
        subschema = schema
        if type(path) != list:
            path = path.split('.')

        for name in path:
            if "properties" in subschema:
                subschema = subschema["properties"]

            subschema = self.get_keyword_value(subschema, name)
            if subschema is None:
                return None

        return subschema

    def get_keyword_value(self, submodel, keyword_name):
        """
        Lppk for keyword name in submodel, return value if found
        """
        keyword_value = None
        if isinstance(submodel, dict):
            if keyword_name in submodel:
                keyword_value = submodel[keyword_name]

            elif "allOf" in submodel:
                for subsubmodel in submodel["allOf"]:
                    if (isinstance(subsubmodel, dict) and
                            keyword_name in subsubmodel):
                        keyword_value = subsubmodel[keyword_name]
                        break

        return keyword_value

    def has_properties(self, model_schema):
        """
        Determine if schema dictionary has any real properties
        """
        properties = self.get_keyword_value(model_schema, "properties")
        if properties is not None:
            model_schema = properties

        return len(model_schema) > 0

    def match_fits_keywords(self, schema_file, model_schema, keyword_schema,
                            keyword_dict, fits_dict, model_path):
        """
        Match model names to keyword db names by fits keyword name
        """
        properties = self.get_keyword_value(model_schema, "properties")
        if properties is not None:
            model_schema = properties

        for model_name in model_schema:
            model_subschema = model_schema[model_name]
            new_model_path = new_path(model_path, model_name)
            if isinstance(model_subschema, dict):
                model_fits_name = self.get_keyword_value(model_subschema,
                                                         "fits_keyword")

                if model_fits_name is not None:
                    # Special case for reference file keywords
                    model_fits_name = model_fits_name.strip()
                    p_name = model_fits_name[0:2] == "P_"
                    if p_name:
                        model_name = model_name[2:]
                        keyword_path = keyword_dict.get(model_fits_name[2:])
                    else:
                        keyword_path = keyword_dict.get(model_fits_name)

                    if keyword_path is None:
                        keyword_subschema = self.find_subschema(keyword_schema,
                                                                new_model_path)
                    else:
                        keyword_subschema = self.find_subschema(keyword_schema,
                                                                keyword_path)

                    if keyword_subschema is not None:
                        keyword_fits_name = keyword_subschema.get("fits_keyword")
                        if keyword_fits_name is not None:
                            fits_dict[model_fits_name] = keyword_fits_name.strip()

                else:
                    fits_hdu = self.get_keyword_value(model_subschema,
                                                      "fits_hdu")
                    if fits_hdu is None:
                        self.match_fits_keywords(schema_file,
                                                 model_subschema,
                                                 keyword_schema,
                                                 keyword_dict,
                                                 fits_dict,
                                                 new_model_path)

            elif isinstance(model_subschema, list) and model_name == "allOf":
                for model_subsubschema in model_subschema:
                    if isinstance(model_subsubschema, dict):
                        self.match_fits_keywords(schema_file,
                                                 model_subsubschema,
                                                 keyword_schema,
                                                 keyword_dict,
                                                 fits_dict,
                                                 new_model_path)

    def pattern_from_enum(self, value):
        """
        Convert the value of an enum field to a regular expression pattern
        """
        return r'^((' + '|'.join(value) + r')\\s*\\|\\s*)+$'

    def report_and_query(self, verb, title, keyword_value, model_value,
                         path, field):
        """
        Report a difference and optionally wait for use input
        """
        if ".".join(path) in self.omit:
            choice = False

        else:
            if self.query or self.list:
                self.report_schema_difference(title,
                                              keyword_value,
                                              model_value,
                                              path,
                                              field)
            if self.query and not self.list:
                choice = self.options.query_user(verb, default_choice=True)
                self.update_omit(choice, path)
            else:
                choice = not self.list

        if choice:
            self.current_file_changed = True

        return choice

    def report_schema_difference(self, title, keyword_value, model_value,
                                 path, field):
        """
        Report the differences between the values of a field in the
        datamodels schema and the keywords database
        """
        self.fd.write("=== {0} ===\n".format('.'.join(path)))
        if title is not None:
            self.fd.write(title + "\n")

        if isinstance(keyword_value, list) and isinstance(model_value, list):
            (keyword_value, model_value) = self.sort_for_report(keyword_value,
                                                                model_value)

        if field == "pattern":
            model_value = self.pattern_from_enum(model_value)
            keyword_value = self.pattern_from_enum(keyword_value)

        self.report_schema_value("model schema", model_value, field)
        self.report_schema_value("keyword db", keyword_value, field)

    def report_schema_value(self, schema_name, schema_value, field):
        """
        Write a message for a single field value
        """

        if field is None:
            if schema_value is None:
                self.fd.write("%s not found\n" % schema_name)
            elif "properties" in schema_value:
                self.fd.write("%s is a group\n" % schema_name)
            else:
                schema_value = schema_value.get('fits_keyword')
                if schema_value is None:
                    self.fd.write("%s not found\n" % schema_name)
                else:
                    schema_value = schema_value.lower()
                    self.fd.write("%s found as %s\n" % (schema_name, schema_value))

        else:
            prefix = "%s %s = " % (schema_name, field)
            if type(schema_value) == list:
                self.fd.write(save_simple_list(prefix, schema_value))
            else:
                self.fd.write(save_scalar(prefix, schema_value))
            if field in ("enum", "pattern"):
                self.fd.write("\n")

    def schema_add_subschema(self, fits_dict, model_schema, keyword_schema,
                             keyword_path):
        """
        Add a subschema tree to the model schema
        """
        self.schema_add_value(model_schema, keyword_schema, keyword_path)

        if "properties" in keyword_schema:
            model_name = keyword_path[-1]
            model_subschema = self.find_subschema(model_schema, model_name)
            keyword_schema = keyword_schema["properties"]

            for keyword_name in keyword_schema:
                keyword_subschema = self.get_keyword_value(keyword_schema,
                                                           keyword_name)

                keyword_fits_name = \
                    self.get_keyword_value(keyword_subschema,
                                           "fits_keyword")

                if keyword_fits_name is None:
                    add = True
                elif keyword_fits_name in fits_dict:
                    add = False
                elif self.builtin_fits_keyword(keyword_fits_name):
                    add = False
                else:
                    add = True

                if add:
                    self.schema_add_subschema(fits_dict,
                                              model_subschema,
                                              keyword_subschema,
                                              new_path(keyword_path,
                                                       keyword_name))

    def schema_add_value(self, model_schema, keyword_schema, keyword_path):
        """
        Add a new keyword to the model schema
        """
        if self.add:
            title = keyword_schema.get('title')
            if self.report_and_query("Add", title, keyword_schema, None,
                                     keyword_path, None):
                keyword_name = keyword_path[-1]
                model_schema[keyword_name] = OrderedDict()

                for keyword_field in ('title', 'type', 'enum', 'default_value',
                                      'fits_keyword'):
                    keyword_value = \
                        self.strip_blanks(keyword_schema.get(keyword_field))
                    if keyword_value is not None and len(keyword_value) > 0:
                        if keyword_field == "default_value":
                            model_field = "default"
                        else:
                            model_field = keyword_field
                        model_schema[keyword_name][model_field] = keyword_value

                if "properties" in keyword_schema:
                    model_schema[keyword_name]["properties"] = OrderedDict()

    def schema_del_value(self, model_schema, model_name, path):
        """
        Delete a keyword from the model schema
        """
        if self.delete:
            title = model_schema.get('title')
            if self.report_and_query("Delete", title, None,
                                     model_schema[model_name], path, None):
                del model_schema[model_name]

    def schema_edit_value(self, model_schema, keyword_value, model_value,
                          path, field):
        """
        Edit a field within the model schema to match the value in the
        keyword database.
        """

        # Logic for performing updates
        perform = (self.edit and
                   keyword_value is not None and
                   model_value is not None)

        if perform:
            title = model_schema.get('title')
            if self.report_and_query("Change", title, keyword_value, model_value,
                                     path, field):
                if field == "pattern":
                    keyword_value = self.pattern_from_enum(keyword_value)
                model_schema[field] = keyword_value

    def schema_rename_value(self, model_schema, keyword_name, model_name, path):
        """
        Rename a field in the model schema to match the keyword database
        """
        if self.rename:
            title = model_schema.get('title')
            if self.report_and_query("Rename", title, keyword_name, model_name,
                                     path, "name"):
                model_schema[keyword_name] = model_schema[model_name]
                del model_schema[model_name]

    def sort_for_report(self, keyword_value, model_value):
        """
        Reorder the values in two lists of values so that the elements that
        differ are first in the list
        """
        common_values = list(set(keyword_value) & set(model_value))

        different_values = list(set(keyword_value) - set(common_values))
        sorted_keyword_value = different_values + common_values

        different_values = list(set(model_value) - set(common_values))
        sorted_model_value = different_values + common_values

        return sorted_keyword_value, sorted_model_value

    def strip_blanks(self, value):
        """
        Remove leading and trailing blanks from values
        """
        def strip_single(elem):
            if isinstance(elem, str):
                elem = elem.strip()
            return elem

        if isinstance(value, list):
            for i in range(len(value)):
                value[i] = strip_single(value[i])
        else:
            value = strip_single(value)

        return value

    def update_omit(self, choice, path):
        """
        Update the omit parameter based on user input
        """
        path = ".".join(path)
        if choice is True:
            self.omit.discard(path)
        elif choice is False:
            self.omit.add(path)

    def update_schema_fields(self, keyword_schema, model_schema, path):
        """
        Compare the fields of a single item between the datamodels schema
        and the keyword database and update when they differ
        """
        p_name = path[-1][0:2] == "p_"
        for model_field in ('fits_keyword', 'type', 'enum', 'pattern', 'default'):
            # Skip checking fits keyword for reference file fields
            if p_name and model_field == 'fits_keyword':
                continue

            # Bridge differences in terminology between
            # keyword db and model schema

            if model_field == "pattern":
                keyword_field = "enum"
            elif model_field == "default":
                keyword_field = "default_value"
            else:
                keyword_field = model_field

            keyword_value = self.strip_blanks(keyword_schema.get(keyword_field))
            model_value = self.strip_blanks(model_schema.get(model_field))

            if model_field == "pattern" and model_value is not None:
                # Pattern is inside innermost set of parentheses
                patstart = model_value.rfind('(') + 1
                patend = model_value.find(')')
                model_value = model_value[patstart:patend].split('|')

            if keyword_value == "float" and model_value == "number":
                keyword_value = "number"

            if keyword_value == "int" and model_value == "integer":
                keyword_value = "integer"

            if keyword_value == "" and model_value is None:
                keyword_value = None

            # Check if value in keyword db differs from value in model schema

            if not self.compare_schema_values(keyword_value, model_value):
                # Check if default value agrees with model schema type
                if model_field == "default":
                    valid = self.check_type(model_schema, keyword_value)
                else:
                    valid = True

                if valid:
                    self.schema_edit_value(model_schema, keyword_value,
                                           model_value, path, model_field)



from .model_base import JwstDataModel


__all__ = ['IRS2Model']


class IRS2Model(JwstDataModel):
    """
    A data model for the IRS2 refpix reference file.

    Parameters
    __________
    irs2_table : numpy table
        Table for IRS2 refpix correction.
        A table with 8 columns and 2916352 (2048 * 712 * 2) rows.  All
        values are float, but these are interpreted as alternating real
        and imaginary parts (real, imag, real, imag, ...) of complex
        values.  There are four columns for ALPHA and four for BETA.

    dq_table : data quality info table
        Table for identifying bad reference pixels.
        A table with three columns (OUTPUT, ODD_EVEN, and MASK) and
        eight rows.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/irs2.schema"



"""Test model resource management"""
import gc

import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal
import psutil

from models import FitsModel


@pytest.mark.parametrize("extension", ["asdf", "fits"])
def test_context_management(extension, tmp_path):
    """Ensure working across context management"""
    original_path = tmp_path / f"original.{extension}"
    new_path = tmp_path / f"new.{extension}"

    process = psutil.Process()
    base_count = len(process.open_files())

    array = np.random.uniform(size=(1024, 1024))

    with FitsModel() as model:
        model.data = array
        model.save(original_path)

    # No files left open by creating and saving a model:
    assert len(process.open_files()) == base_count

    with FitsModel(original_path) as model:
        # Count should be higher due to opening the file:
        model_count = len(process.open_files())
        assert model_count > base_count

        new_model = FitsModel(model)

    # No files should be closed yet because new_model still
    # needs the resources:
    assert len(process.open_files()) == model_count

    new_model.save(new_path)
    new_model.close()

    # Back to the original state:
    assert len(process.open_files()) == base_count

    # Confirm that the array was written accurately:
    with FitsModel(new_path) as model:
        assert_array_almost_equal(model.data, array)


@pytest.mark.parametrize("extension", ["asdf", "fits"])
def test_close(extension, tmp_path):
    """Ensure file resources exist after re-instantiation
    After re-instantiating a model that has been constructed
    from a file, the original model should be closeable
    without affecting the underlying file resources for the new
    model.
    """
    original_path = tmp_path / f"original.{extension}"
    new_path = tmp_path / f"new.{extension}"

    process = psutil.Process()
    base_count = len(process.open_files())

    array = np.random.uniform(size=(1024, 1024))

    model = FitsModel()
    model.data = array
    model.save(original_path)
    model.close()

    # No files left open by creating and saving a model:
    assert len(process.open_files()) == base_count

    model = FitsModel(original_path)
    # Count should be higher due to opening the file:
    model_count = len(process.open_files())
    assert model_count > base_count

    new_model = FitsModel(model)
    model.close()

    # No files should be closed yet because new_model still
    # needs the resources:
    assert len(process.open_files()) == model_count

    new_model.save(new_path)
    new_model.close()

    # Back to the original state:
    assert len(process.open_files()) == base_count

    # Confirm that the array was written accurately:
    model = FitsModel(new_path)
    assert_array_almost_equal(model.data, array)
    model.close()


@pytest.mark.parametrize("extension", ["asdf", "fits"])
def test_multiple_close(extension, tmp_path):
    """
    Confirm that multiple calls to close() on the same
    model does not prematurely close the file.
    """
    file_path = tmp_path / f"test.{extension}"

    process = psutil.Process()
    base_count = len(process.open_files())

    array = np.random.uniform(size=(1024, 1024))

    model = FitsModel()
    model.data = array
    model.save(file_path)
    model.close()

    # No files left open by creating and saving a model:
    assert len(process.open_files()) == base_count

    model = FitsModel(file_path)
    # Count should be higher due to opening the file:
    model_count = len(process.open_files())
    assert model_count > base_count

    new_model = FitsModel(model)
    model.close()

    # No files should be closed yet because new_model still
    # needs the resources:
    assert len(process.open_files()) == model_count

    # Close the original model a second time:
    model.close()

    # Still no files closed:
    assert len(process.open_files()) == model_count

    new_model.close()

    # Back to the original state:
    assert len(process.open_files()) == base_count


@pytest.mark.parametrize("extension", ["asdf", "fits"])
def test_delete(extension, tmp_path):
    """Deleting the model should also not close files
    until the last model has been deleted."""
    original_path = tmp_path / f"original.{extension}"
    new_path = tmp_path / f"new.{extension}"

    process = psutil.Process()
    base_count = len(process.open_files())

    array = np.random.uniform(size=(1024, 1024))

    model = FitsModel()
    model.data = array
    model.save(original_path)
    del model
    gc.collect()

    # No files left open by creating and saving a model:
    assert len(process.open_files()) == base_count

    model = FitsModel(original_path)
    # Count should be higher due to opening the file:
    model_count = len(process.open_files())
    assert model_count > base_count

    new_model = FitsModel(model)
    del model
    gc.collect()

    # No files should be closed yet because new_model still
    # needs the resources:
    assert len(process.open_files()) == model_count

    new_model.save(new_path)
    del new_model
    gc.collect()

    # Back to the original state:
    assert len(process.open_files()) == base_count

    # Confirm that the array was written accurately:
    model = FitsModel(new_path)
    assert_array_almost_equal(model.data, array)
    del model
    gc.collect()



from .model_base import JwstDataModel
from .image import ImageModel


__all__ = ['IFUImageModel']


class IFUImageModel(JwstDataModel):
    """
    A data model for 2D IFU images.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    zeroframe : numpy float32 array
         Zeroframe array

    var_poisson : numpy float32 array
         variance due to poisson noise

    var_rnoise : numpy float32 array
         variance due to read noise

    wavelength : numpy float32 array
         wavelength

    pathloss_point : numpy float32 array
         pathloss correction for point source

    pathloss_uniform : numpy float32 array
         pathloss correction for uniform source

    area : numpy float32 array
         Pixel area map array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ifuimage.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, ImageModel):
            super(IFUImageModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.data = init.data
            self.dq = init.dq
            self.err = init.err
            if init.hasattr('var_poisson'):
                self.var_poisson = init.var_poisson
            if init.hasattr('var_rnoise'):
                self.var_rnoise = init.var_rnoise
            if init.hasattr('area'):
                self.area = init.area
            if init.hasattr('pathloss_point'):
                self.pathloss_point = init.pathloss_point
            if init.hasattr('pathloss_uniform'):
                self.pathloss_uniform = init.pathloss_uniform
            return

        super(IFUImageModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



from .reference import ReferenceFileModel
from .extract1dimage import Extract1dImageModel


__all__ = ['MultiExtract1dImageModel']


class MultiExtract1dImageModel(ReferenceFileModel):
    """
    A data model for extract_1d reference images.

    This model has a special member `images` that can be used to
    deal with each image separately.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import Extract1dImageModel
       >>> multiextr1d_img_model = MultiExtract1dImageModel()
       >>> multiextr1d_img_model.images.append(Extract1dImageModel())
       >>> multiextr1d_img_model.images[0]  # doctest: +SKIP
       <Extract1dImageModelModel>

    If `init` is a file name or an `Extract1dImageModel` instance, an empty
    `Extract1dImageModel` will be created and assigned to attribute
    `images[0]`, and the `data` attribute from the input array or
    `Extract1dImageModel` will be copied to the first element of `images`.

    Parameters
    __________
    images.items.data : numpy float32 array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/multiextract1d.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, Extract1dImageModel):
            super(MultiExtract1dImageModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.images.append(self.images.item())
            self.images[0].data = init.data
            return

        super(MultiExtract1dImageModel, self).__init__(init=init, **kwargs)



from pathlib import Path
import os
import sys
from configparser import ConfigParser
from datetime import datetime
import importlib

import stsci_rtd_theme
if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

def setup(app):
    try:
        app.add_css_file("stsci.css")
    except AttributeError:
        app.add_stylesheet("stsci.css")


REPO_ROOT = Path(__file__).parent.parent.parent

# Modules that automodapi will document need to be available
# in the path:
sys.path.insert(0, str(REPO_ROOT/"src"/"stdatamodels"))

with open(REPO_ROOT / "pyproject.toml", 'rb') as configuration_file:
    setup_metadata = tomllib.load(configuration_file)['project']

project = setup_metadata["name"]
author = setup_metadata["authors"][0]["name"]
copyright = f"{datetime.now().year}, {author}"

package = importlib.import_module(setup_metadata["name"])
version = package.__version__.split("-", 1)[0]
release = package.__version__

extensions = [
    "sphinx_automodapi.automodapi",
    "numpydoc",
    'sphinx_asdf',
]

autosummary_generate = True
numpydoc_show_class_members = False
autoclass_content = "both"

html_theme = "stsci_rtd_theme"
html_theme_options = {
    "collapse_navigation": True
}
html_theme_path = [stsci_rtd_theme.get_html_theme_path()]
html_domain_indices = True
html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}
html_use_index = True



from .reference import ReferenceFileModel

__all__ = ['NirspecIFUCubeParsModel', 'MiriIFUCubeParsModel']


class NirspecIFUCubeParsModel(ReferenceFileModel):
    """
    A data model for Nirspec ifucubepars reference files.

    Parameters
    __________
    ifucubepars_table : numpy table
         default IFU cube  parameters table

    ifucubepars_msm_table : numpy table
         default IFU cube msm parameters table

    ifucubepars_prism_msm_wavetable : numpy table
         default IFU cube prism msm wavetable

    ifucubepars_med_msm_wavetable : numpy table
         default IFU cube med resolution msm  wavetable

    ifucubepars_high_msm_wavetable : numpy table
         default IFU cube high resolution msm wavetable

    ifucubepars_emsm_table : numpy table
         default IFU cube emsm parameters table

    ifucubepars_prism_emsm_wavetable : numpy table
         default IFU cube prism emsm wavetable

    ifucubepars_med_emsm_wavetable : numpy table
         default IFU cube med resolution emsm wavetable

    ifucubepars_high_emsm_wavetable : numpy table
         default IFU cube high resolution emsm  wavetable

    ifucubepars_prism_driz_wavetable : numpy float32 array
         default IFU cube prism drizzle wavetable

    ifucubepars_med_driz_wavetable : numpy float32 array
         default IFU cube med resolution drizzle wavetable

    ifucubepars_high_driz_wavetable : numpy float32 array
         default IFU cube high resolution drizzle wavetable
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_ifucubepars.schema"


class MiriIFUCubeParsModel(ReferenceFileModel):
    """
    A data model for MIRI mrs ifucubepars reference files.

    Parameters
    __________
    ifucubepars_table : numpy table
         default IFU cube  parameters table

    ifucubepars_msm_table : numpy table
         default IFU cube msm parameters table

    ifucubepars_multichannel_msm_wavetable : numpy table
         default IFU cube msm wavetable

    ifucubepars_emsm_table : numpy table
         default IFU cube emsm parameters table

    ifucubepars_multichannel_emsm_wavetable : numpy table
         default IFU cube emsm wavetable

    ifucubepars_multichannel_driz_wavetable : numpy float32 array
         default IFU cube driz wavetable

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/miri_ifucubepars.schema"



from .model_base import JwstDataModel

__all__ = ['AmiOIModel']


class AmiOIModel(JwstDataModel):
    """
    TODO

    Parameters
    __________
    TODO
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/amioi.schema"

    def get_primary_array_name(self):
        # for the example file OI_T3 is the largest array
        return 't3'

    def on_save(self, path=None):
        super().on_save(path)

        # OIFITS requires specific keyword names for some data that already
        # exist in JWST files under different keywords. For existing
        # (and defined) metadata, copy the data to the OIFITS compatible
        # locations

        self.meta.oifits.derived.observer = self.meta.program.pi_name
        self.meta.oifits.derived.object = self.meta.target.proposer_name or self.meta.target.catalog_name

        # This file contains OIFITS2 content
        self.meta.oifits.derived.content = 'OIFITS2'

        # each OIFITS data table needs specific cross-referencing keywords
        array_name = self.meta.oifits.array_name
        self.meta.oifits.derived.t3.array_name = array_name
        self.meta.oifits.derived.vis.array_name = array_name
        self.meta.oifits.derived.vis2.array_name = array_name

        insname = self.meta.instrument.name
        self.meta.oifits.derived.wavelength.instrument_name = insname
        self.meta.oifits.derived.t3.instrument_name = insname
        self.meta.oifits.derived.vis.instrument_name = insname
        self.meta.oifits.derived.vis2.instrument_name = insname

        # JWST saves meta.observation.date to DATE-OBS
        # which is the UTC date for the observation start
        # In Table 2 of the OIFITS 2 paper:
        # https://doi.org/10.48550/arXiv.1510.04556
        # However it's unclear is this format is required
        # and if a UTC date is valid.
        # Currently this code leaves the value as-is (expressed
        # as a date only) to avoid complications for other JWST
        # code.
        date_obs = self.meta.observation.date
        self.meta.oifits.derived.t3.date_obs = date_obs
        self.meta.oifits.derived.vis.date_obs = date_obs
        self.meta.oifits.derived.vis2.date_obs = date_obs

        self.meta.oifits.derived.array.revn = 2
        self.meta.oifits.derived.target.revn = 2
        self.meta.oifits.derived.t3.revn = 2
        self.meta.oifits.derived.vis.revn = 2
        self.meta.oifits.derived.vis2.revn = 2
        self.meta.oifits.derived.wavelength.revn = 2

        # fill in possibly missing OI_ARRAY meta data
        if self.meta.oifits.derived.array.frame is None:
            self.meta.oifits.derived.array.frame = 'SKY'
        if self.meta.oifits.derived.array.x is None:
            self.meta.oifits.derived.array.x = 0.0
        if self.meta.oifits.derived.array.y is None:
            self.meta.oifits.derived.array.y = 0.0
        if self.meta.oifits.derived.array.z is None:
            self.meta.oifits.derived.array.z = 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from asdf_astropy.converters.transform.core import TransformConverterBase
from astropy.modeling import Model
from ....properties import ListNode

__all__ = ['Gwa2SlitConverter', 'Slit2MsaConverter', 'LogicalConverter', 'NirissSOSSConverter',
           'RefractionIndexConverter', 'MIRI_AB2SliceConverter', 'NIRCAMGrismDispersionConverter',
           'NIRISSGrismDispersionConverter', 'Rotation3DToGWAConverter', 'GratingEquationConverter',
           'V23ToSkyConverter', 'SnellConverter', 'CoordsConverter', 'Rotation3DToGWAConverter']


class NIRCAMGrismDispersionConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/nircam_grism_dispersion-*"]

    types = ["stdatamodels.jwst.transforms.models.NIRCAMForwardRowGrismDispersion",
             "stdatamodels.jwst.transforms.models.NIRCAMForwardColumnGrismDispersion",
             "stdatamodels.jwst.transforms.models.NIRCAMBackwardGrismDispersion"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms import models

        # Current version uses model_type
        if 'model_type' in node.keys():
            _fname = getattr(models, node["model_type"])

            return _fname(node.get('orders'),
                          node.get('lmodels'),
                          node.get('xmodels'),
                          node.get('ymodels'),
                          )

        # Backwards-compatible class_name, with no saved inverse models
        elif 'class_name'in node.keys():
            _fname = getattr(models, node["class_name"])

            return _fname(list(node['orders']),
                          list(node['lmodels']),
                          list(node['xmodels']),
                          list(node['ymodels']),
                          )
        else:
            raise KeyError("YAML tree did not have expected node key of 'model_type' or 'class_name'.")

    def to_yaml_tree_transform(self, model, tag, ctx):
        # Second order modeling has list of lists of models
        if isinstance(model.xmodels[0], Model):
            xll = list(model.xmodels)
            yll = list(model.ymodels)
            lll = list(model.lmodels)

        # First order modeling has list of models
        elif isinstance(model.xmodels[0], (ListNode, list)):
            xll = [list(m) for m in model.xmodels]
            yll = [list(m) for m in model.ymodels]
            lll = [list(m) for m in model.lmodels]
        else:
            raise KeyError(f"xmodels entry is not a Model or an iterable. Type: {type(model.xmodels[0])}")

        node = {'orders': list(model.orders),
                'xmodels': xll,
                'ymodels': yll,
                'lmodels': lll,
                'model_type': type(model).name
                }
        return node


class NIRISSGrismDispersionConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/niriss_grism_dispersion-*"]

    types = ["stdatamodels.jwst.transforms.models.NIRISSForwardRowGrismDispersion",
             "stdatamodels.jwst.transforms.models.NIRISSForwardColumnGrismDispersion",
             "stdatamodels.jwst.transforms.models.NIRISSBackwardGrismDispersion"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms import models

        _fname = getattr(models, node["model_type"])
        return _fname(list(node['orders']),
                      list(node['lmodels']),
                      list(node['xmodels']),
                      list(node['ymodels']),
                      node['theta'],
                      )

    def to_yaml_tree_transform(self, model, tag, ctx):
        xll = [list(m) for m in model.xmodels]
        yll = [list(m) for m in model.ymodels]
        node = {'orders': list(model.orders),
                'xmodels': xll,
                'ymodels': yll,
                'lmodels': list(model.lmodels),
                'theta': model.theta,
                'model_type': type(model).name
                }
        return node


class Gwa2SlitConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/gwa_to_slit-*"]

    types = ["stdatamodels.jwst.transforms.models.Gwa2Slit"]

    def from_yaml_tree_transform(self, node, tag, ctx):

        from stdatamodels.jwst.transforms.models import Gwa2Slit

        return Gwa2Slit(node['slits'], node['models'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'slits': model._slits,
                'models': model.models}
        return node


class Slit2MsaConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/slit_to_msa-*"]

    types = ["stdatamodels.jwst.transforms.models.Slit2Msa"]

    def from_yaml_tree_transform(self, node, tag, ctx):

        from stdatamodels.jwst.transforms.models import Slit2Msa

        return Slit2Msa(node['slits'], node['models'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'slits': model._slits,
                'models': model.models}
        return node


class LogicalConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/logical-*"]
    types = ["stdatamodels.jwst.transforms.models.Logical"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import Logical

        return Logical(node['condition'], node['compareto'], node['value'])

    def to_yaml_tree_transform(self, model, tag, ctx):

        node = {'condition': model.condition,
                'compareto': model.compareto,
                'value': model.value}
        return node


class NirissSOSSConverter(TransformConverterBase):
    tags = ["tag:stsci.edu:jwst_pipeline/niriss-soss-*",
            "tag:stsci.edu:jwst_pipeline/niriss_soss-*"]
    types = ["stdatamodels.jwst.transforms.models.NirissSOSSModel"]

    def from_yaml_tree_transform(self, node, tag, ctx):

        from stdatamodels.jwst.transforms.models import NirissSOSSModel

        return NirissSOSSModel(node['spectral_orders'], node['models'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'spectral_orders': list(model.models.keys()),
                'models': list(model.models.values())
                }
        return node


class RefractionIndexConverter(TransformConverterBase):
    tags = ["tag:stsci.edu:jwst_pipeline/refraction_index_from_prism-*"]
    types = ["stdatamodels.jwst.transforms.models.RefractionIndexFromPrism"]

    def from_yaml_tree_transform(self, node, tag, ctx):

        from stdatamodels.jwst.transforms.models import RefractionIndexFromPrism

        return RefractionIndexFromPrism(node['prism_angle'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'prism_angle': model.prism_angle.value
                }
        return node


class MIRI_AB2SliceConverter(TransformConverterBase):
    tags = ["tag:stsci.edu:jwst_pipeline/miri_ab2slice-*"]
    types = ["stdatamodels.jwst.transforms.models.MIRI_AB2Slice"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import MIRI_AB2Slice
        return MIRI_AB2Slice(node['beta_zero'], node['beta_del'], node['channel'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'beta_zero': model.beta_zero.value,
                'beta_del': model.beta_del.value,
                'channel': model.channel.value
                }
        return node


class Rotation3DToGWAConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/rotation_sequence-*"]

    types = ["stdatamodels.jwst.transforms.models.Rotation3DToGWA"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import Rotation3DToGWA

        angles = node['angles']
        axes_order = node['axes_order']
        return Rotation3DToGWA(angles, axes_order)

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'angles': list(model.angles.value)}
        node['axes_order'] = model.axes_order
        return node


class GratingEquationConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/grating_equation-*"]

    types = ["stdatamodels.jwst.transforms.models.WavelengthFromGratingEquation",
             "stdatamodels.jwst.transforms.models.AngleFromGratingEquation"
             ]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import WavelengthFromGratingEquation, AngleFromGratingEquation

        groove_density = node['groove_density']
        order = node['order']
        output = node['output']
        if output == "wavelength":
            model = WavelengthFromGratingEquation(groove_density, order)
        elif output == "angle":
            model = AngleFromGratingEquation(groove_density, order)
        else:
            raise ValueError("Can't create a GratingEquation model with"
                             "output {0}".format(output))
        return model

    def to_yaml_tree_transform(self, model, tag, ctx):
        from stdatamodels.jwst.transforms.models import WavelengthFromGratingEquation, AngleFromGratingEquation

        node = {'order': model.order.value,
                'groove_density': model.groove_density.value
                }
        if isinstance(model, AngleFromGratingEquation):
            node['output'] = 'angle'
        elif isinstance(model, WavelengthFromGratingEquation):
            node['output'] = 'wavelength'
        else:
            raise TypeError("Can't serialize an instance of {0}"
                            .format(model.__class__.__name__))
        return node


class V23ToSkyConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/v23tosky-*"]

    types = ["stdatamodels.jwst.transforms.models.V23ToSky"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import V23ToSky
        angles = node['angles']
        axes_order = node['axes_order']
        return V23ToSky(angles, axes_order=axes_order)

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'angles': list(model.angles.value)}
        node['axes_order'] = model.axes_order
        return node


class SnellConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/snell-*"]

    types = ["stdatamodels.jwst.transforms.models.Snell"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import Snell
        return Snell(node['prism_angle'], node['kcoef'], node['lcoef'], node['tcoef'],
                     node['ref_temp'], node['ref_pressure'], node['temp'], node['pressure'])

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {'prism_angle': model.prism_angle,
                'kcoef': model.kcoef.tolist(),
                'lcoef': model.lcoef.tolist(),
                'tcoef': model.tcoef.tolist(),
                'ref_temp': model.tref,
                'ref_pressure': model.pref,
                'temp': model.temp,
                'pressure': model.pressure
                }
        return node


class CoordsConverter(TransformConverterBase):

    tags = ["tag:stsci.edu:jwst_pipeline/coords-*"]

    types = ["stdatamodels.jwst.transforms.models.Unitless2DirCos",
             "stdatamodels.jwst.transforms.models.DirCos2Unitless"]

    def from_yaml_tree_transform(self, node, tag, ctx):
        from stdatamodels.jwst.transforms.models import Unitless2DirCos, DirCos2Unitless

        model_type = node['model_type']
        if model_type in ('to_dircos', 'unitless2directional'):
            return Unitless2DirCos()
        elif model_type in ('from_dircos', 'directional2unitless'):
            return DirCos2Unitless()
        else:
            raise TypeError("Unknown model_type")

    def to_yaml_tree_transform(self, model, tag, ctx):
        from stdatamodels.jwst.transforms.models import Unitless2DirCos, DirCos2Unitless

        if isinstance(model, DirCos2Unitless):
            model_type = 'directional2unitless'
        elif isinstance(model, Unitless2DirCos):
            model_type = 'unitless2directional'
        else:
            raise TypeError("Model of type {0} i snot supported."
                            .format(model.__class__))
        node = {'model_type': model_type}
        return node



from .reference import ReferenceFileModel


__all__ = ['PathlossModel', 'MirLrsPathlossModel']


class PathlossModel(ReferenceFileModel):
    """
    A data model for pathloss correction information.

    Parameters
    __________
    apertures.items.pointsource_data : numpy float32 array
         Point source pathloss

    apertures.items.pointsource_err : numpy float32 array
         Point source pathloss variance

    apertures.items.uniform_data : numpy float32 array
         Uniform source pathloss

    apertures.items.uniform_err : numpy float32 array
         Uniform source pathloss variance
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/pathloss.schema"


class MirLrsPathlossModel(ReferenceFileModel):
    """
    A data model for MIRI LRS pathloss correction information.

    Parameters
    __________
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/mirlrs_pathloss.schema"



from .model_base import JwstDataModel
from .combinedspec import CombinedSpecModel


__all__ = ['MultiCombinedSpecModel']


class MultiCombinedSpecModel(JwstDataModel):
    """
    A data model for multi-spec images.

    This model has a special member `spec` that can be used to
    deal with an entire spectrum at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import CombinedSpecModel
       >>> multispec_model = MultiCombinedSpecModel()
       >>> multispec_model.spec.append(CombinedSpecModel())
       >>> multispec_model.spec[0] # doctest: +SKIP
       <CombinedSpecModel>

    If `init` is a `CombinedSpecModel` instance, an empty `CombinedSpecModel`
    will be created and assigned to attribute `spec[0]`, and the `spec_table`
    attribute from the input `CombinedSpecModel` instance will be copied to
    the first element of `spec`.  `CombinedSpecModel` objects can be appended
    to the `spec` attribute by using its `append` method.

    Parameters
    __________
    int_times : numpy table
         table of times for each integration

    spec.items.spec_table : numpy table
         Extracted spectral data table

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/multicombinedspec.schema"

    def __init__(self, init=None, **kwargs):

        if isinstance(init, CombinedSpecModel):
            super(MultiCombinedSpecModel, self).__init__(init=None, **kwargs)
            self.spec.append(self.spec.item())
            self.spec[0].spec_table = init.spec_table
            return

        super(MultiCombinedSpecModel, self).__init__(init=init, **kwargs)



from datetime import datetime, timedelta

import pytest

import asdf
from astropy.io import fits
import numpy as np
from numpy.testing import assert_array_equal
from asdf.tags.core import Software, HistoryEntry

from stdatamodels import util


def test_gentle_asarray_array_input():
    """
    Test gentle_asarray with various forms of np.ndarray input,
    which may need to be cast to a different data type.
    """
    # Array that already bears the correct dtype:
    inp = np.array([1, 2, 3, 4], dtype=np.int16)
    result = util.gentle_asarray(inp, dtype=np.int16)
    assert result.dtype == np.int16
    assert_array_equal(result, inp)

    # This will require casting, since int8 != int16:
    result = util.gentle_asarray(inp, dtype=np.int8)
    assert result.dtype == np.int8
    assert_array_equal(result, inp)

    # Test broadcasting the same 1D array to a recarray with
    # two columns.  The input array should be copied into
    # each column and cast as necessary:
    dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    result = util.gentle_asarray(inp, dtype=dtype)
    assert result.dtype == dtype
    assert result["col1"].dtype == np.int16
    assert_array_equal(result["col1"], np.array([1, 2, 3, 4], dtype=np.int16))
    assert result["col2"].dtype == np.float32
    assert_array_equal(result["col1"], np.array([1, 2, 3, 4], dtype=np.float32))


def test_gentle_asarray_recarray_input():
    """
    Test gentle_asarray input that is already an np.recarray.
    """
    # This is just the simple case where the dtype already matches
    # the desired output, but it follows a different code path
    # from array input:
    dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=dtype)
    result = util.gentle_asarray(inp, dtype=dtype)
    assert result.dtype == dtype
    assert_array_equal(result, inp)


def test_gentle_asarray_fits_rec_input():
    """
    Test gentle_asarray with FITS_rec array input.
    """
    # 'e' is the FITS data type code for single-precision float:
    cols = [fits.Column("col1", format="e", array=np.array([1, 2, 3, 4]))]
    inp = fits.FITS_rec.from_columns(cols)
    out_dtype = np.dtype([("col1", np.float32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype
    assert_array_equal(result, inp)


def test_gentle_asarray_fits_rec_pseudo_unsigned(tmp_path):
    """
    Test gentle_asarray handling of a FITS_rec with a pseudo unsigned
    integer column, which is a special case due to a bug in astropy.
    """
    # 'j' is the FITS data type code for 32-bit integer.  A column
    # is only considered "pseudo unsigned" for certain values of
    # bezero, one of which is 2**31.
    cols = [fits.Column("col1", format="j", array=np.array([1, 2, 3, 4], np.uint32), bzero=2**31)]
    inp = fits.FITS_rec.from_columns(cols)
    out_dtype = np.dtype([("col1", np.uint32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    # If not handled, the astropy bug would result in a signed
    # int column dtype:
    assert result["col1"].dtype == np.uint32
    assert result["col1"][0] == 1
    assert result["col1"][1] == 2
    assert result["col1"][2] == 3
    assert result["col1"][3] == 4

    # This tests the case where a table with a pseudo unsigned integer column
    # is opened from a FITS file and needs to be cast.  This requires special
    # handling on our end to dodge the bug.
    file_path = tmp_path / "test.fits"

    data = np.array([(0,)], dtype=[("col1", np.uint16)])
    hdu = fits.BinTableHDU()
    hdu.data = data
    hdul = fits.HDUList([fits.PrimaryHDU(), hdu])
    hdul.writeto(file_path)

    with fits.open(file_path) as hdul:
        result = util.gentle_asarray(hdul[-1].data, dtype=[("col1", np.uint32)])
        # Without the fix, the value in the array would be 128 due to bzero
        # shift being applied twice.
        assert result[0][0] == 0


def test_gentle_asarray_nested_array():
    """
    Test gentle_asarray with nested arrays in one of the record
    fields.
    """
    # '2f' is a numpy data type code for an array of two 32-bit floats:
    in_dtype = np.dtype([("col1", np.dtype("2f")), ("col2", np.int16)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("col1", np.dtype("2f")), ("col2", np.int8)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype
    assert_array_equal(result["col2"], inp["col2"])


def test_gentle_asarray_mismatched_field_names():
    """
    Test gentle_asarray with input field names that don't
    match the requested output field names.
    """
    in_dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("col1", np.int16), ("foo", np.float32)])
    with pytest.raises(ValueError):
        util.gentle_asarray(inp, dtype=out_dtype)


def test_gentle_asarray_field_name_case():
    """
    Test gentle_assary with input and output field names that
    only differ by case.
    """
    in_dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("COL1", np.int16), ("COL2", np.float32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype


def test_gentle_asarray_scalar_input():
    """
    Test gentle_asarray with scalar input.
    """
    result = util.gentle_asarray(3.14159, dtype=np.float32)
    assert result.dtype == np.float32
    assert_array_equal(result, np.array([3.14159], dtype=np.float32))


def test_gentle_asarray_invalid_conversion():
    """
    Test gentle_asarray with input that cannot be cast
    to an array.
    """
    with pytest.raises(ValueError):
        util.gentle_asarray(object(), dtype=np.float32)


def test_create_history_entry():
    entry = util.create_history_entry("Once upon a time...")
    assert isinstance(entry, HistoryEntry)
    assert entry["description"] == "Once upon a time..."
    assert entry.get("software") is None
    assert (datetime.utcnow() - entry["time"]) < timedelta(seconds=10)

    software = {"name": "PolarBearSoft", "version": "1.2.3"}
    entry = util.create_history_entry("There was a tie-dyed polar bear...", software)
    assert isinstance(entry["software"], Software)
    assert entry["software"]["name"] == "PolarBearSoft"
    assert entry["software"]["version"] == "1.2.3"

    software = [
        {"name": "PolarBearSoft", "version": "1.2.3"},
        {"name": "BanjoSoft", "version": "4.5.6"},
    ]
    entry = util.create_history_entry("Who loved to strum the banjo.", software)
    assert isinstance(entry["software"], list)
    assert all(isinstance(s, Software) for s in entry["software"])
    assert entry["software"][0]["name"] == "PolarBearSoft"
    assert entry["software"][0]["version"] == "1.2.3"
    assert entry["software"][1]["name"] == "BanjoSoft"
    assert entry["software"][1]["version"] == "4.5.6"


@pytest.mark.parametrize(
    "value,expected_result",
    [
        ("0", False),
        ("false", False),
        ("FALSE", False),
        ("f", False),
        ("no", False),
        ("n", False),
        ("1", True),
        ("-1", True),
        ("198815238", True),
        ("true", True),
        ("TRUE", True),
        ("t", True),
        ("yes", True),
        ("y", True),
    ]
)
def test_get_envar_as_boolean(monkeypatch, value, expected_result):
    monkeypatch.setenv("TEST_VAR", value)
    assert util.get_envar_as_boolean("TEST_VAR") is expected_result


def test_get_envar_as_boolean_default(monkeypatch):
    monkeypatch.delenv("TEST_VAR", raising=False)
    assert util.get_envar_as_boolean("TEST_VAR") is False
    assert util.get_envar_as_boolean("TEST_VAR", default=False) is False
    assert util.get_envar_as_boolean("TEST_VAR", default=True) is True


def test_get_envar_as_boolean_invalid(monkeypatch):
    monkeypatch.setenv("TEST_VAR", "foo")
    with pytest.raises(ValueError):
        assert util.get_envar_as_boolean("TEST_VAR")


def test_get_model_type():
    assert util.get_model_type(asdf.AsdfFile()) is None
    assert util.get_model_type(asdf.AsdfFile({"meta": {"model_type": "SomeModel"}})) == "SomeModel"

    assert util.get_model_type(fits.HDUList([fits.PrimaryHDU()])) is None

    hdu = fits.PrimaryHDU()
    hdu.header["DATAMODL"] = "SomeOtherModel"
    assert util.get_model_type(fits.HDUList([hdu])) == "SomeOtherModel"



from .reference import ReferenceFileModel


__all__ = ['SpecKernelModel']


class SpecKernelModel(ReferenceFileModel):
    """
    A data model for 2D spectral kernels.

    Parameters
    __________
    wavelengths : numpy float32 array
         Wavelengths

    kernels : numpy float32 array
         Kernel values
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/speckernel.schema"



from .model_base import JwstDataModel

__all__ = ['SossWaveGridModel']


class SossWaveGridModel(JwstDataModel):
    """
    A data model to hold NIRISS SOSS wavelength grids.
    This 1-D array of wavelengths can be saved from a processing run
    and applied to future input products.

    Parameters
    __________
    wavegrid : numpy float32 array
        1-D array of the wavelengths corresponding to the ATOCA fit.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/sosswavegrid.schema"



from copy import deepcopy

from asdf import schema as asdf_schema
from asdf import treeutil

from .model_base import JwstDataModel
from .image import ImageModel
from .slit import SlitModel, SlitDataModel


__all__ = ['MultiExposureModel', 'set_hdu', 'remove_fits']


class MultiExposureModel(JwstDataModel):
    """
    A data model for multi-slit images derived from
    numerous exposures. The intent is that all slits
    in this model are of the same source, with each slit
    representing a separate exposure of that source.

    This model has a special member `exposures` that can be used to
    deal with an entire slit at a time.  It behaves like a list::

       >>> from .image import ImageModel
       >>> multiexposure_model = MultiExposureModel()
       >>> multiexposure_model.exposures.append(ImageModel())
       >>> multiexposure_model.exposures[0]      # doctest: +SKIP
       <ImageModel>

    Also, there is an extra attribute, `meta`. This will contain the
    meta attribute from the exposure from which each slit has been taken.

    See the module `exp_to_source` for the initial creation of these
    models. This is part of the Level 3 processing of multi-objection
    observations.

    Parameters
    __________
    exposures.items.data : numpy float32 array

    exposures.items.dq : numpy uint32 array

    exposures.items.err : numpy float32 array

    exposures.items.area : numpy float32 array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/multiexposure.schema"
    core_schema_url = 'http://stsci.edu/schemas/jwst_datamodel/core.schema'

    def __init__(self, init=None, **kwargs):

        # Lets create a schema
        schema = self._build_schema()

        if isinstance(init, (SlitModel, SlitDataModel, ImageModel)):
            super(MultiExposureModel, self).__init__(
                init=None,
                schema=schema,
                **kwargs
            )
            self.update(init)
            self.exposures.append(self.exposures.item())
            self.exposures[0].data = init.data
            self.exposures[0].dq = init.dq
            self.exposures[0].err = init.err
            self.exposures[0].var_poisson = init.var_poisson
            self.exposures[0].var_rnoise = init.var_rnoise
            self.exposures[0].var_flat = init.var_flat
            self.exposures[0].wavelength = init.wavelength
            self.exposures[0].barshadow = init.barshadow
            self.exposures[0].flatfield_point = init.flatfield_point
            self.exposures[0].flatfield_uniform = init.flatfield_uniform
            self.exposures[0].pathloss_point = init.pathloss_point
            self.exposures[0].pathloss_uniform = init.pathloss_uniform
            self.exposures[0].photom_point = init.photom_point
            self.exposures[0].photom_uniform = init.photom_uniform
            self.exposures[0].area = init.area
            return

        super(MultiExposureModel, self).__init__(
            init=init,
            schema=schema,
            **kwargs
        )

    def _build_schema(self):
        """Build the schema, incorporating the core."""
        # Get the schemas
        schema = asdf_schema.load_schema(
            self.schema_url,
            resolve_references=True
        )
        core_schema = asdf_schema.load_schema(
            self.core_schema_url,
            resolve_references=True
        )

        # Create a new core.meta that will co-locate
        # with each exposure entry. This is done
        # by saving the meta information in a separate
        # FITS HDU.
        core_meta_schema = deepcopy(core_schema['properties']['meta'])
        treeutil.walk(core_meta_schema, remove_fits)
        exposure_schema = schema['allOf'][1]['properties']['exposures']['items']
        exposure_meta_schema = exposure_schema['allOf'][1]['properties']['meta']
        exposure_meta_schema.update(core_meta_schema)

        # That's all folks
        return schema


# Utilities
def set_hdu(obj, hdu_id='EXP'):
    """Add fits_hdu specification to fits-connected properties"""
    try:
        if 'fits_keyword' in obj.keys():
            obj['fits_hdu'] = hdu_id
    except AttributeError:
        pass


def remove_fits(obj):
    try:
        obj.pop('fits_keyword', None)
    except (AttributeError, KeyError, TypeError):
        pass
    try:
        obj.pop('fits_hdu', None)
    except (AttributeError, KeyError, TypeError):
        pass



import warnings
from astropy.time import Time

from stdatamodels import DataModel as _DataModel


__all__ = ["DataModel", "JwstDataModel"]


class JwstDataModel(_DataModel):

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/core.schema"

    @property
    def crds_observatory(self):
        """
        Get CRDS observatory code for this model.

        Returns
        -------
        str
        """
        return "jwst"

    def get_crds_parameters(self):
        """
        Get parameters used by CRDS to select references for this model.

        Returns
        -------
        dict
        """
        return {
            key: val for key, val in self.to_flat_dict(include_arrays=False).items()
            if isinstance(val, (str, int, float, complex, bool))
        }

    def on_init(self, init):
        """
        Hook invoked by the base class before returning a newly
        created model instance.
        """
        super().on_init(init)

        if not self.meta.hasattr("date"):
            self.meta.date = Time.now().isot

    def on_save(self, init):
        """
        Hook invoked by the base class before writing a model
        to a file (FITS or ASDF).
        """
        super().on_save(init)

        self.meta.date = Time.now().isot


class DataModel(JwstDataModel):
    def __init_subclass__(cls):
        warnings.warn("Class DataModel has been renamed JwstDataModel. It will be removed in v 1.5.", DeprecationWarning, 2)



from .reference import ReferenceFileModel


__all__ = ['PixelAreaModel', 'NirspecSlitAreaModel', 'NirspecMosAreaModel',
           'NirspecIfuAreaModel']


class PixelAreaModel(ReferenceFileModel):
    """
    A data model for the pixel area map

    Parameters
    __________
    data : numpy float32 array
         The pixel area array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/pixelarea.schema"


class NirspecSlitAreaModel(ReferenceFileModel):
    """
    A data model for the NIRSpec fixed-slit pixel area reference file

    Parameters
    __________
    area_table : numpy table
         NIRSpec fixed-slit pixel area table
         A table-like object containing row selection criteria made up
         of the slit id and the pixel area values associated with
         the slits.

         - slit_id: str[15]
         - pixarea: float32
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_area_slit.schema"


class NirspecMosAreaModel(ReferenceFileModel):
    """
    A data model for the NIRSpec MOS pixel area reference file

    Parameters
    __________
    area_table : numpy table
         NIRSpec MOS pixel area table
         A table-like object containing row selection criteria made up
         of MOS shutter parameters and the pixel area values associated
         with the shutters.

         - quadrant: int16
         - shutter_x: int16
         - shutter_y: int16
         - pixarea: float32
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_area_mos.schema"


class NirspecIfuAreaModel(ReferenceFileModel):
    """
    A data model for the NIRSpec IFU pixel area reference file

    Parameters
    __________
    area_table : numpy table
         NIRSpec IFU pixel area table
         A table-like object containing row selection criteria made up
         of IFU slice id and the pixel area values associated
         with the slices.

         - slice_id: int16
         - pixarea: float32
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_area_ifu.schema"



from stdatamodels.dynamicdq import dynamic_mask
from .dqflags import pixel
from .reference import ReferenceFileModel


__all__ = ['LinearityModel']


class LinearityModel(ReferenceFileModel):
    """
    A data model for linearity correction information.

    Parameters
    __________
    coeffs : numpy float32 array
         Linearity coefficients

    dq : numpy uint32 array
         Data quality flags

    dq_def : numpy table
         DQ flag definitions
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/linearity.schema"

    def __init__(self, init=None, **kwargs):
        super(LinearityModel, self).__init__(init=init, **kwargs)

        self.dq = dynamic_mask(self, pixel)

        # Implicitly create arrays
        self.dq = self.dq

    def get_primary_array_name(self):
        """
        Returns the name "primary" array for this model, which
        controls the size of other arrays that are implicitly created.
        This is intended to be overridden in the subclasses if the
        primary array's name is not "data".
        """
        return 'coeffs'



from .reference import ReferenceFileModel

__all__ = ['BarshadowModel']


class BarshadowModel(ReferenceFileModel):
    """
    A data model for Bar Shadow correction information.

    Parameters
    __________
    data1x1 : numpy float32 array
         Bar Shadow 1x1 data array

    var1x1 : numpy float32 array
         Bar Shadow 1x1 correction variance

    data1x3 : numpy float32 array
         Bar Shadow 1x3 data array

    var1x3 : numpy float32 array
         Bar Shadow 1x3 correction variance
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/barshadow.schema"



""" JWST Data Quality Flags

The definitions are documented in the JWST RTD:

https://jwst-pipeline.readthedocs.io/en/latest/jwst/references_general/references_general.html#data-quality-flags


Implementation
-------------

The flags are implemented as "bit flags": Each flag is assigned a bit position
in a byte, or multi-byte word, of memory. If that bit is set, the flag assigned
to that bit is interpreted as being set or active.

The data structure that stores bit flags is just the standard Python `int`,
which provides 32 bits. Bits of an integer are most easily referred to using
the formula `2**bit_number` where `bit_number` is the 0-index bit of interest.
"""

# These imports are here for backwards compatibility
from astropy.nddata.bitmask import interpret_bit_flags as ap_interpret_bit_flags
from stdatamodels.dqflags import interpret_bit_flags, dqflags_to_mnemonics
from stdatamodels.basic_utils import multiple_replace

# Pixel-specific flags
pixel = {'GOOD':             0,      # No bits set, all is good
         'DO_NOT_USE':       2**0,   # Bad pixel. Do not use.
         'SATURATED':        2**1,   # Pixel saturated during exposure
         'JUMP_DET':         2**2,   # Jump detected during exposure
         'DROPOUT':          2**3,   # Data lost in transmission
         'OUTLIER':          2**4,   # Flagged by outlier detection (was RESERVED_1)
         'PERSISTENCE':      2**5,   # High persistence (was RESERVED_2)
         'AD_FLOOR':         2**6,   # Below A/D floor (0 DN, was RESERVED_3)
         'UNDERSAMP':        2**7,   # Undersampling correction (was RESERVED_4)
         'UNRELIABLE_ERROR': 2**8,   # Uncertainty exceeds quoted error
         'NON_SCIENCE':      2**9,   # Pixel not on science portion of detector
         'DEAD':             2**10,  # Dead pixel
         'HOT':              2**11,  # Hot pixel
         'WARM':             2**12,  # Warm pixel
         'LOW_QE':           2**13,  # Low quantum efficiency
         'RC':               2**14,  # RC pixel
         'TELEGRAPH':        2**15,  # Telegraph pixel
         'NONLINEAR':        2**16,  # Pixel highly nonlinear
         'BAD_REF_PIXEL':    2**17,  # Reference pixel cannot be used
         'NO_FLAT_FIELD':    2**18,  # Flat field cannot be measured
         'NO_GAIN_VALUE':    2**19,  # Gain cannot be measured
         'NO_LIN_CORR':      2**20,  # Linearity correction not available
         'NO_SAT_CHECK':     2**21,  # Saturation check not available
         'UNRELIABLE_BIAS':  2**22,  # Bias variance large
         'UNRELIABLE_DARK':  2**23,  # Dark variance large
         'UNRELIABLE_SLOPE': 2**24,  # Slope variance large (i.e., noisy pixel)
         'UNRELIABLE_FLAT':  2**25,  # Flat variance large
         'OPEN':             2**26,  # Open pixel (counts move to adjacent pixels)
         'ADJ_OPEN':         2**27,  # Adjacent to open pixel
         'FLUX_ESTIMATED':   2**28,  # Pixel flux estimated due to missing/bad data
         'MSA_FAILED_OPEN':  2**29,  # Pixel sees light from failed-open shutter
         'OTHER_BAD_PIXEL':  2**30,  # A catch-all flag
         'REFERENCE_PIXEL':  2**31,  # Pixel is a reference pixel
         }


# Group-specific flags. Once groups are combined, these flags
# are equivalent to the pixel-specific flags.
group = {'GOOD':       pixel['GOOD'],
         'DO_NOT_USE': pixel['DO_NOT_USE'],
         'SATURATED':  pixel['SATURATED'],
         'JUMP_DET':   pixel['JUMP_DET'],
         'DROPOUT':    pixel['DROPOUT'],
         'AD_FLOOR':   pixel['AD_FLOOR'],
         'UNDERSAMP':  pixel['UNDERSAMP'],
         }

__all__ = ["ap_interpret_bit_flags", "interpret_bit_flags", "dqflags_to_mnemonics",
           "multiple_replace", "pixel", "group"]



from io import BytesIO

import asdf
from asdf.constants import ASDF_MAGIC
import numpy as np
from numpy.testing import assert_array_equal
from astropy.io import fits

from models import FitsModel


_NDARRAY_TAG = "tag:stsci.edu:asdf/core/ndarray-1.0.0"


def create_fits_model():
    data = np.asarray(np.random.rand(50, 50), np.float32)
    dq = np.random.randint(np.iinfo(np.uint32).max, size=(50, 50), dtype=np.uint32)
    err = np.asarray(np.random.rand(50, 50), np.float32)

    model = FitsModel((50, 50))
    model["data"] = data
    model["dq"] = dq
    model["err"] = err

    return model, data, dq, err


def open_embedded_asdf(file_path):
    with fits.open(file_path) as hdul:
        return asdf.open(
            BytesIO(hdul["ASDF"].data["ASDF_METADATA"].tobytes()),
            _force_raw_types=True,
        )


def test_asdf_hdu_format(tmp_path):
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    model.save(file_path)

    with fits.open(file_path) as hdul:
        # One for the primary HDU, three for the arrays, and one for the ASDF HDU
        assert len(hdul) == 5

        # ASDF-in-FITS should be saved to a binary table HDU named "ASDF"
        # with a single uint8 column named "ASDF_METADATA".
        assert "ASDF" in hdul
        asdf_hdu = hdul["ASDF"]
        assert isinstance(asdf_hdu, fits.hdu.table.BinTableHDU)
        assert asdf_hdu.data.dtype.type == np.record
        assert len(asdf_hdu.data.dtype) == 1
        assert asdf_hdu.data.dtype.names == ("ASDF_METADATA",)
        assert asdf_hdu.data.dtype[0].base == np.uint8
        assert asdf_hdu.data.dtype[0].ndim == 1

        # Does it smell like an ASDF file?
        asdf_bytes = asdf_hdu.data["ASDF_METADATA"].tobytes()
        assert asdf_bytes.startswith(ASDF_MAGIC)

        # "assert" no exceptions when opening the content as ASDF.
        # Force raw types to avoid errors when the ndarray converter
        # encounters the linked FITS arrays.
        fd = BytesIO(asdf_bytes)
        asdf.open(fd, _force_raw_types=True)


def test_linked_arrays(tmp_path):
    file_path = tmp_path / "test.fits"

    model, data, dq, err = create_fits_model()
    model.save(file_path)

    af = open_embedded_asdf(file_path)

    assert "data" in af
    tagged_data = af["data"]
    assert tagged_data._tag == _NDARRAY_TAG
    assert tagged_data["source"] == "fits:SCI,1"
    assert tagged_data["datatype"] == "float32"
    assert tagged_data["byteorder"] == "big"
    assert tagged_data["shape"] == [50, 50]

    assert "dq" in af
    tagged_dq = af["dq"]
    assert tagged_dq._tag == _NDARRAY_TAG
    assert tagged_dq["source"] == "fits:DQ,1"
    assert tagged_dq["datatype"] == "uint32"
    assert tagged_dq["byteorder"] == "big"
    assert tagged_dq["shape"] == [50, 50]

    assert "err" in af
    tagged_err = af["err"]
    assert tagged_err._tag == _NDARRAY_TAG
    assert tagged_err["source"] == "fits:ERR,1"
    assert tagged_err["datatype"] == "float32"
    assert tagged_err["byteorder"] == "big"
    assert tagged_err["shape"] == [50, 50]

    with fits.open(file_path) as hdul:
        with FitsModel(hdul) as dm:
            assert_array_equal(dm.data, data)
            # If linking wasn't working correctly, these would be different objects:
            assert dm.data is hdul["SCI"].data

            assert_array_equal(dm.dq, dq)
            assert dm.dq is hdul["DQ"].data

            assert_array_equal(dm.err, err)
            assert dm.err is hdul["ERR"].data


def test_non_fits_array(tmp_path):
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    favorite_integers = np.random.randint(np.iinfo(np.uint32).max, size=(500,), dtype=np.uint32)
    model["favorite_integers"] = favorite_integers
    model.save(file_path)

    with fits.open(file_path) as hdul:
        # We shouldn't have gained an HDU:
        assert len(hdul) == 5

    af = open_embedded_asdf(file_path)
    assert "favorite_integers" in af
    # Should be an integer internal block source and not a string FITS source:
    assert isinstance(af["favorite_integers"]["source"], int)

    with FitsModel(file_path) as dm:
        assert_array_equal(dm.favorite_integers, favorite_integers)


def test_non_fits_metadata(tmp_path):
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    model["splines"] = "reticulated"
    model.save(file_path)

    with fits.open(file_path) as hdul:
        assert "splines" not in hdul[0].header

    af = open_embedded_asdf(file_path)
    assert af["splines"] == "reticulated"

    with FitsModel(file_path) as dm:
        assert dm.splines == "reticulated"


def test_array_update_and_save_new_file(tmp_path):
    file_path = tmp_path / "test.fits"
    updated_file_path = tmp_path / "updated.fits"

    model, _, _, _ = create_fits_model()
    model.save(file_path)

    new_data = np.asarray(np.random.rand(50, 50), np.float32)
    with FitsModel(file_path) as dm:
        dm.data = new_data
        dm.save(updated_file_path)

    af = open_embedded_asdf(updated_file_path)
    assert af["data"]["source"] == "fits:SCI,1"

    with fits.open(updated_file_path) as hdul:
        with FitsModel(hdul) as dm:
            assert_array_equal(dm.data, new_data)
            assert dm.data is hdul["SCI"].data


def test_array_update_and_save_same_file(tmp_path):
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    model.save(file_path)

    new_data = np.asarray(np.random.rand(50, 50), np.float32)
    with FitsModel(file_path) as dm:
        dm.data = new_data
        dm.save(file_path)

    af = open_embedded_asdf(file_path)
    assert af["data"]["source"] == "fits:SCI,1"

    with fits.open(file_path) as hdul:
        with FitsModel(hdul) as dm:
            assert_array_equal(dm.data, new_data)
            assert dm.data is hdul["SCI"].data


def test_fits_array_view(tmp_path):
    """
    The old ASDF-in-FITS support provided by the asdf library
    was able to write views over FITS arrays to the embedded
    ASDF as views.  JWST doesn't make use of that feature, so
    this package doesn't bother and instead writes the view
    as a separate array in the ASDF tree.
    """
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    view = model.data[:25, :]
    model["view"] = view
    model.save(file_path)

    af = open_embedded_asdf(file_path)
    assert isinstance(af["view"]["source"], int)

    with fits.open(file_path) as hdul:
        # We shouldn't have gained an HDU:
        assert len(hdul) == 5

    with FitsModel(file_path) as dm:
        assert_array_equal(dm["view"], view)


def test_hdu_link_independence(tmp_path):
    """
    As links between arrays and hdu items are made during
    saving, it's possible that if this goes wrong links
    might be made between multiple arrays and a single hdu.
    In this case, modifying one array will change memory
    shared with another array. This test creates a file
    with multiple arrays, writes it to a fits file,
    reads it back in and then modifies the contents of
    each array to check for this possible error.
    """
    file_path = tmp_path / "test.fits"

    model, _, _, _ = create_fits_model()
    model.save(file_path)

    with FitsModel(file_path) as dm:
        # assign new values
        dm.data[:] = 1
        dm.dq[:] = 2
        dm.err[:] = 3

        assert np.all(dm.data == 1)
        assert np.all(dm.dq == 2)
        assert np.all(dm.err == 3)


def test_resave_breaks_hdulist_tree_array_link(tmp_path):
    """
    Test that writing, reading and rewriting an AsdfInFits file
    maintains links between hdus and arrays in the asdf tree

    If the link is broken, data can be duplicated (exist both
    as a hdu and as an internal block in the asdf tree).

    See issues:
        https://github.com/asdf-format/asdf/issues/1232
        https://github.com/spacetelescope/jwst/issues/7354
        https://github.com/spacetelescope/jwst/issues/7274
    """
    file_path_1 = tmp_path / "test1.fits"
    file_path_2 = tmp_path / "test2.fits"

    model, _, _, _ = create_fits_model()
    model.save(file_path_1)

    with FitsModel(file_path_1) as dm:
        dm.save(file_path_2)

    # check that af1 (original write) and af2 (rewrite) do not contain internal ASDF blocks
    with fits.open(file_path_1) as af1, fits.open(file_path_2) as af2:
        for f in (af1, af2):
            block_bytes = f["ASDF"].data.tobytes().split(b"...")[1].strip()
            assert len(block_bytes) == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re
from collections import OrderedDict


__all__ = ["find_fits_keyword", "search_schema", "walk_schema",
           "merge_property_trees", "build_docstring", "SearchSchemaResults"]


# return_result included for backward compatibility
def find_fits_keyword(schema, keyword, return_result=False):
    """
    Utility function to find a reference to a FITS keyword in a given
    schema.  This is intended for interactive use, and not for use
    within library code.

    Parameters
    ----------
    schema : JSON schema fragment
        The schema in which to search.

    keyword : str
        A FITS keyword name

    Returns
    -------
    locations : list of str
    """
    def find_fits_keyword(subschema, path, combiner, ctx, recurse):
        if len(path) and path[0] == 'extra_fits':
            return True
        if subschema.get('fits_keyword') == keyword:
            results.append('.'.join(path))

    results = []
    walk_schema(schema, find_fits_keyword, results)

    return results


class SearchSchemaResults(list):
    def __repr__(self):
        import textwrap

        result = []
        for path, description in self:
            result.append(path)
            result.append(
                textwrap.fill(
                    description, initial_indent='    ',
                    subsequent_indent='    '))
        return '\n'.join(result)


def search_schema(schema, substring):
    """
    Utility function to search the metadata schema for a particular
    phrase.

    This is intended for interactive use, and not for use within
    library code.

    The searching is case insensitive.

    Parameters
    ----------
    schema : JSON schema fragment
        The schema in which to search.

    substring : str
        The substring to search for.

    Returns
    -------
    locations : list of tuples
    """
    substring = substring.lower()

    def find_substring(subschema, path, combiner, ctx, recurse):
        matches = False
        for param in ('title', 'description'):
            if substring in schema.get(param, '').lower():
                matches = True
                break

        if substring in '.'.join(path).lower():
            matches = True

        if matches:
            description = '\n\n'.join([
                schema.get('title', ''),
                schema.get('description', '')]).strip()
            results.append(('.'.join(path), description))

    results = SearchSchemaResults()
    walk_schema(schema, find_substring, results)
    results.sort()
    return results


def walk_schema(schema, callback, ctx=None):
    """
    Walks a JSON schema tree in breadth-first order, calling a
    callback function at each entry.

    Parameters
    ----------
    schema : JSON schema

    callback : callable
        The callback receives the following arguments at each entry:

        - subschema: The subschema for the entry
        - path: A list of strings defining the path to the entry
        - combiner: The current combiner in effect, will be 'allOf',
          'anyOf', 'oneOf', 'not' or None
        - ctx: An arbitrary context object, usually a dictionary
        - recurse: A function to call to recurse deeper on a node.

        If the callback returns `True`, the subschema will not be
        further recursed.

    ctx : object, optional
        An arbitrary context object
    """
    def recurse(schema, path, combiner, ctx):
        if callback(schema, path, combiner, ctx, recurse):
            return

        for c in ['allOf', 'not']:
            for sub in schema.get(c, []):
                recurse(sub, path, c, ctx)

        for c in ['anyOf', 'oneOf']:
            for i, sub in enumerate(schema.get(c, [])):
                recurse(sub, path + [c], c, ctx)

        if schema.get('type') == 'object':
            for key, val in schema.get('properties', {}).items():
                recurse(val, path + [key], combiner, ctx)

        if schema.get('type') == 'array':
            items = schema.get('items', {})
            if isinstance(items, list):
                for i, item in enumerate(items):
                    recurse(item, path + [i], combiner, ctx)
            elif len(items):
                recurse(items, path + ['items'], combiner, ctx)

    if ctx is None:
        ctx = {}
    recurse(schema, [], None, ctx)


def merge_property_trees(schema):
    """
    Recursively merges property trees that are governed by the "allOf" combiner.

    The main purpose of this function is to allow multiple subschemas to be
    combined into a single schema. All of the properties at each level of each
    subschema are merged together to form a single coherent tree.

    This allows datamodel schemas to be more modular, since various components
    can be represented in individual files and then referenced elsewhere. They
    are then combined by this function into a single schema data structure.
    """
    newschema = OrderedDict()

    def add_entry(path, schema, combiner):
        # TODO: Simplify?
        cursor = newschema
        for i in range(len(path)):
            part = path[i]
            if part == combiner:
                cursor = cursor.setdefault(combiner, [])
                return
            elif isinstance(part, int):
                cursor = cursor.setdefault('items', [])
                while len(cursor) <= part:
                    cursor.append({})
                cursor = cursor[part]
            elif part == 'items':
                cursor = cursor.setdefault('items', OrderedDict())
            else:
                cursor = cursor.setdefault('properties', OrderedDict())
                if i < len(path) - 1 and isinstance(path[i + 1], int):
                    cursor = cursor.setdefault(part, [])
                else:
                    cursor = cursor.setdefault(part, OrderedDict())

        cursor.update(schema)

    def callback(schema, path, combiner, ctx, recurse):
        type = schema.get('type')
        schema = OrderedDict(schema)
        if type == 'object':
            del schema['properties']
        elif type == 'array':
            del schema['items']
        if 'allOf' in schema:
            del schema['allOf']

        add_entry(path, schema, combiner)

    walk_schema(schema, callback)

    return newschema


def build_docstring(klass, template="{fits_hdu} {title}"):
    """
    Build a docstring for the specified DataModel class from its schema.


    Parameters
    ----------

    klass : Python class
        A class instance of a datamodel
    template : str
        A string format template to be applied to each schema item

    Returns
    -------
    field_info : str
        Information about each schema item associated with a FITS hdu
    """
    from . import model_base

    def get_field_info(subschema, path, combiner, info, recurse):
        # Return all schema fields representing fits hdus
        if 'fits_hdu' in subschema and 'fits_keyword' not in subschema:
            attr = '.'.join(path)
            info[attr] = subschema
        return 'fits_hdu' in subschema or 'fits_keyword' in subschema

    # Silly rabbit, only datamodels have schemas
    if not (klass == model_base.DataModel or
            issubclass(klass, model_base.DataModel)):
        raise ValueError("Class must be a subclass of DataModel: %s",
                         klass.__name__)

    # Create a new model just to get its shape
    null_object = klass(init=None)
    shape = null_object.shape
    if shape is None:
        shaped_object = null_object
    else:
        # Instantiate an object with correctly dimensioned shape
        null_object.close()
        shape = tuple([1 for i in range(len(shape))])
        shaped_object = klass(init=shape)

    # Get schema fields which have an associated hdu
    info = {}
    walk_schema(shaped_object._schema, get_field_info, ctx=info)

    # Extract field names from template to set defaults
    # so format won't crash while using them when they aren't there
    default_schema = {}
    fields = re.findall(r'\{([^\\:}]*)[\:\}]', template)
    for field in fields:
        default_schema[field] = ''

    buffer = []
    for attr, subschema in info.items():
        schema = {}
        schema.update(default_schema)
        schema.update(subschema)
        schema['path'] = attr

        # Determine if attribute has a default value
        instance = shaped_object.instance
        for field in attr.split('.'):
            try:
                instance = instance.get(field)
            except AttributeError:
                instance = None
            if instance is None:
                break
        schema['default'] = instance is not None

        # Extract table field names from datatype
        if type(schema['datatype']) == str:
            schema['array'] = True
        else:
            schema['records'] = True
            fields = []
            for field_info in schema['datatype']:
                fields.append(field_info['name'])
            schema['fields'] = ', '.join(fields)
            schema['datatype'] = 'table'

        # Convert boolean fields to their field names
        for field, value in schema.items():
            if type(value) == bool:
                schema[field] = field

        # Apply format to schema fields
        # Delete blank lines
        lines = template.format(**schema)
        for line in lines.split("\n"):
            if line and not line.isspace():
                buffer.append(line)

    field_info = "\n".join(buffer) + "\n"
    return field_info



from .reference import ReferenceFileModel

__all__ = ['ResolutionModel', 'MiriResolutionModel']


class ResolutionModel(ReferenceFileModel):
    """
    A data model for Spectral Resolution  parameters reference tables.

    Parameters
    __________
    data : numpy float32 array
        Resolving Power table
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/resolution.schema"


class MiriResolutionModel(ResolutionModel):
    """
    A data model for MIRI Resolution reference files.

    Parameters
    __________
    data : numpy table
        Resolving Power table
        A table containing resolving power of the MRS. THe table consist of 11
        columns and 12 rows. Each row corresponds to a band. The columns give the
        name of band, central wavelength, and polynomial coefficients (a,b,c)
        needed to obtain the limits and average value of the spectral resolution.

    psf_fwhm_alpha_table : table
        PSF FWHM Alpha
        A table with 5 columns. Column 1 gives the cutoff wavelength where the
        polynomials describing alpha FWHM change. Columns 2 and 3 give the
        polynomial coefficients (a,b) describing alpha FWHM for wavelengths
        shorter than cutoff. Columns 4 and 5 give the polynomial
        coefficients (a,b) describing alpha FWHM for wavelengths longer than the
        cutoff.

    psf_fwhm_beta_table : table
        PSF FWHM Beta
        A table with 5 columns. Column 1 gives the cutoff wavelength where the
        polynomials describing alpha FWHM change. Columns 2 and 3 give the
        polynomial coefficients (a,b) describing beta FWHM for wavelengths shorter
        than cutoff. Columns 4 and 5 give the polynomial coefficients (a,b)
        describing beta FWHM for wavelengths longer than the cutoff.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/miri_resolution.schema"



from .model_base import JwstDataModel


__all__ = ['SpecModel']


class SpecModel(JwstDataModel):
    """
    A data model for 1D spectra.

    Parameters
    __________
    spec_table : numpy table
        Extracted spectral data table
        A table with at least four columns:  wavelength, flux, an error
        estimate for the flux, and data quality flags.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/spec.schema"



"""
This module supports the entry points for ASDF support for the `jwst.transforms`.
"""

import sys

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


from asdf.resource import DirectoryResourceMapping
from stdatamodels.jwst import transforms


def get_resource_mappings():
    """
    Get the `jwst.transforms` resource mappings, that is the schemas for the transforms.

    This method is registered with the `asdf.resource_mappings` entry point for
    the `jwst_pipeline`.

    Returns
    -------
    list of the `asdf.resource.ResourceMapping` instances containing the `jwst.transforms`
    schemas.
    """
    resources_root = importlib_resources.files(transforms) / "resources"
    if not resources_root.is_dir():
        raise RuntimeError(f"Missing resources directory: {resources_root=}")

    return [
        DirectoryResourceMapping(
            resources_root / "schemas" / "stsci.edu" / "jwst_pipeline",
            "http://stsci.edu/schemas/jwst_pipeline/",
        ),
        DirectoryResourceMapping(
            resources_root / "manifests",
            "asdf://stsci.edu/jwst_pipeline/manifests/",
        ),
    ]


def get_extensions():
    """
    Get the jwst.transforms extension.
    This method is registered with the asdf.extensions entry point.

    Returns
    -------
    list of asdf.extension.Extension
    """
    from . import extensions
    return extensions.TRANSFORM_EXTENSIONS
